/*=============================================================================
 *
 * MODULE
 *
 *     LfapCCEos.c
 *   
 * ABSTRACT
 *    Implement LFAP API Interfaces to Nucleus OS
 *
 *    Cabletron Systems Incorporated
 *    Post Office Box 5005
 *    Rochester, NH 03866-5005
 *    (c) Copyright Cabletron Systems Inc. 1997
 *
 ===========================================================================*/

/******************************************************************************/
/*                                                                            */
/*  File: LFAPCCEos.c                                                         */
/*                                                                            */
/*  This file is for debugging purposes only.  It is used for simulating      */
/*  the actions of a switch using the LFAP API.  The functions in this file   */
/*  'wrap back' to the LFAP API, calling the appropriate LFAP API calls.  The */
/*  functions send back sample data when necessary.                           */
/*                                                                            */
/******************************************************************************/

/* 
 * WARNING WARNING
 *
 * since there are currently no clients on a windows platform
 * and since we don't want this file included when the windows
 * server is built under Visual C++, put in this ifdef
*/
#ifndef _WIN32


/* Start of auto-generated includes */
#define INCLUDES_H_
#ifdef MLS_SOL_BUILD
#include "sol_includes.h"
#endif
#ifdef MLS_MIPS_BUILD
#include "cdefs.h"
#include "nucleus.h"
#include "ni_api.h" 
#include "mls_in.h"
#include "mls_resolv.h"
#include "mls_netdb.h"
#include "mls_reent.h"
#include "mls_stdio.h"
#include "mls_stdlib.h"
#include "mls_errno.h"
#include "mls_string.h"
#include "mls_select.h"
#include "byteorder.h"
#endif
#include "lfap_acl.h"
#include "l3port_api.h"
#include "rtc_api.h"
#include "uio.h"
#include "mls_socket.h"
#include "mls_inet.h"
#include "mls_ioctl.h"
#include "mls_syscalls.h"
#include "err_api.h"
/* End of auto-generated includes */

#include "LfapCCEos.h"
#include "LfapFlowMgr.h"
#include "lfap_interface.h"
#include "lfap.emd.h"
#define INCL_LFAP_STATE
#include "lfap_main.h"

int vec_r = 0;                  /* vectors remaining for LFAPSendVector() */
extern NU_TASK LFAP_TCB;        /* in tasks.c */
extern NU_TASK* lfapWorkerThread; 
int fas_sd = -1;                /* socket descriptor to fas */
static int lfap_socket_full = 0;/* indicates socket is not writable */
/* maximum msg we would ever receive from FAS
 */
#define MAX_LFAP_MSG_SZ (1024)

extern int lfap_debug;
extern int lfap_efacid;
extern void lfap_connection_lost(unsigned long server);
extern void lfap_connection_established(unsigned long server);

/* 
 * pre: lfap socket has data msg to read
 * post: return's length of buffer including header
 * NOTES: return blocks on on read based on msg len
 * msg fmt: HDR_LEN(datalen) | data
 */
int get_lfap_msg(char *buf, int *len)
{
    int rc = 0;
    char* ptr = buf;
    int act_len = 0;
    int l = LFAP_HEADER_SIZE;

    /* 1. get msg header (spin till header is read) */
    while (l > 0) {
        rc = read(fas_sd, ptr, l);
        if (rc == 0) {
            /* connection was lost */
            return 0;
        }
        else if (rc == -1) {
            if (l < LFAP_HEADER_SIZE) {
                /* partial header read */
                switch (errno) {
                  case EAGAIN:
                  case EINTR:
                    NU_Sleep(1);
                    break;
                  default:
                    return 0;
                }
            }
            else {
                /* nothing read */
                switch (errno) {
                  case EINTR:
                    continue;
                    break;
                  case EAGAIN:
                  default:
                    return -1;
                }
            }
        }
        else {
            ptr += rc;
            l -= rc;
        }
    }

    /* 2. decode header to get msg length */
    act_len = LFAPDecodeMsgLen(buf);
    
    /* 3. validate length */
    if (act_len > MAX_LFAP_MSG_SZ) {
        #ifdef CS_DEBUG
        if (lfap_debug & LFAP_DBG_ERRORS) {
            ERR_errlog(0, lfap_efacid, LFAP_E_MSG_TOO_LARGE, 1, act_len);
        }
        #endif
        return -1;
    }
    
    /* 4. read rest of body (spins till all data is read) */
    if (act_len > 0) {
        l = act_len;
        ptr = buf + LFAP_HEADER_SIZE;
        while (l > 0) {
            rc = read(fas_sd, ptr, l);
            if (rc == 0) {
                /* connection was lost */
                return 0;
             }
            else if (rc == -1) {
                /* partial message read */
                switch (errno) {
                  case EAGAIN:
                  case EINTR:
                    NU_Sleep(1);
                    break;
                  default:
                    return 0;
                }
            }
            else {
                ptr += rc;
                l -= rc;
            }
        }
    }

    *len = act_len + LFAP_HEADER_SIZE;
    return act_len + LFAP_HEADER_SIZE;
}


/******************************************************************************/
/*                                                                            */
/*  Function: LFAPGetUptime                                                   */
/*                                                                            */
/*  This function is called by the LFAP API to get the current time.      */
/*                                                                            */
/*  return a time value in 100th' seconds                                     */
/******************************************************************************/
lfapui32_t LFAPGetCurTime (void)
{
    return time(NULL);
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPGetSysUptime                                                   */
/*                                                                            */
/*  This function is called by the LFAP API to get the CCE uptime value.      */
/*                                                                            */
/*  return a time value in 100th' seconds                                     */
/******************************************************************************/
lfapui32_t LFAPGetSysUptime (void)
{
    return NUE_CurrTimeTicks();
}


/******************************************************************************/
/*                                                                            */
/*  Function: LFAPGetStatistics                                               */
/*                                                                            */
/*  This function is called by the LFAP API to get the statistics on a flow   */
/*  with identifier flow_id.  A pointer to a zero initialized                 */
/*  LFAPFlowStats structure is passed in to be filled in by this              */
/*  function.  Note that this function should be autonomous (no contact       */
/*  switches if on a threaded operating system).                              */
/*                                                                            */
/******************************************************************************/
void LFAPGetStatistics (lfapui32_t flow_id, LFAPFlowStats* flow_stats)
{
    lfap_get_statistics(flow_id, flow_stats);
}


/******************************************************************************
 *                                                                            *
 *  Function: LFAPDoesFlowExist                                               *
 *                                                                            *
 *  This function is called by the LFAP API to determine if the flow still    *
 *  exists in the CCE environment. Since the flow may have been deleted by    *
 *  the hardware some time ago, care must be take to implement this function  *
 *  in such a way that time is not a factor (i.e. if hardware flow handles    *
 *  are reused, this function should not depend on them                       *
 *                                                                            *
 * NOTE - The main user is anyone who uses flows from LFAP_stats_list.        *
 *        since flows are deleted from the hardware but                       *
 *        are not immediately deleted from the LFAP_stats_list                *
 *        (the garabage collector does the actual delete)                     *
 *        we need to make sure the flow is still valid before                 *
 *        we try to get stats on it.                                          *
 *                                                                            *
 ******************************************************************************/
int LFAPDoesFlowExist (lfapui32_t flow_id)
{
    return lfap_does_flow_exist(flow_id);
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPGetSrcAS                                                    */
/*                                                                            */
/*  return the Autonomous System (AS) number for the given address.           */
/*                                                                            */
/******************************************************************************/
lfapui32_t LFAPGetSrcAS (lfapui32_t src_addr)
{
    return lfap_get_src_as(src_addr);
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPSynchronize                                                 */
/*                                                                            */
/*  This function is called by the LFAP API to allow a non-preemptive system  */
/*  to run other tasks.  The CCE should complete whatever work it wishes to   */
/*  complete, then return from this function.  Returning from this function   */
/*  will allow the LFAP API to continue with its work.  This function should  */
/*  (at a minimum) allow a context switch to occur.                           */
/*                                                                            */
/******************************************************************************/
void LFAPSynchronize (void)
{
    NU_Sleep(100);                   /* try for context switch (units: ticks)*/
}

/* pre: timer expired and nucleus invoked this routine
 * post: LFAP task is signalled to process the timer.
 */
typedef void (*expiration_routine)(UNSIGNED); /* hack/need hdr for this */

static void
lfap_callback(unsigned val)
{
  int status;

  if (lfap_timers[val].status == LFAP_TIMER_WAIT) {
      lfap_timers[val].status = LFAP_TIMER_SET;
  }
  else if (lfap_timers[val].status == LFAP_TIMER_DISABLED) {
      lfap_timers[val].status = LFAP_TIMER_DISABLED_SET;
  }

  if ((status = NUE_Set_TEG_Flag(&LFAP_TCB, LFAP_TIMER_EXPIRED_BIT))
      != NU_SUCCESS) {
      #ifdef CS_DEBUG
      if (lfap_debug & LFAP_DBG_ERRORS)
          ERR_errlog(0, lfap_efacid, LFAP_E_CALLBACK_FAIL, 0);
      #endif
  }
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPSetSystemTimer                                              */
/*                                                                            */
/*  This function is called by the LFAP API to set a timer for time_val       */
/*  milliseconds.  lfap_cb_entry is the callback entrypoint that is called    */
/*  with the parameter return_handle upon completion of the timer.            */
/*                                                                            */
/******************************************************************************/
void LFAPSetSystemTimer (const LFAP_CB_ENTRYPOINT lfap_cb_entry,
                         lfapui32_t return_handle, lfapui32_t time_val)
{
  int timer;
  int status;
  char buf[20];
  lfapui32_t time_ticks;
  static unsigned current_timer = 0;
  static int all_timers_created = 0;

  time_ticks = time_val * RTC_TICKSIN1SEC / 1000;

  if (all_timers_created < LFAP_MAX_TIMERS_OUTSTANDING) {

      #ifdef CS_DEBUG
      if (lfap_debug & LFAP_DBG_TIMERS)
      {
      fprintf(stderr,"SetSystemTimer: registering timer: %u \n", current_timer);
      fprintf(stderr, "             setting for: %lu milliseconds\n", time_val);
      fprintf(stderr, "             setting for: %lu time ticks\n", time_ticks);
      }
      #endif

      lfap_timers[current_timer].status = LFAP_TIMER_WAIT;
      lfap_timers[current_timer].cb = (LFAP_CB_ENTRYPOINT) lfap_cb_entry;
      lfap_timers[current_timer].handle = return_handle;
      lfap_timers[current_timer].wait_time = time_val;

      sprintf(buf, "LFAP_TIMER_%u", current_timer);
      status = NU_Create_Timer(&lfap_timers[current_timer].timer, buf,
                               (expiration_routine)lfap_callback, 
                               current_timer,
                               time_ticks, /* initial timeo */
                               0, /* reschedule every */
                               NU_ENABLE_TIMER);
      if (status != NU_SUCCESS) {
          #ifdef CS_DEBUG
          if (lfap_debug & LFAP_DBG_ERRORS)
              ERR_errlog(0, lfap_efacid, LFAP_E_TIMER_FAIL, 1, strerror(errno));
          #endif
      }
      else {
          all_timers_created++;
          if (++current_timer == LFAP_MAX_TIMERS_OUTSTANDING) {
              current_timer = 0;
          }
      }
  }
  else {

      if (lfap_timers[current_timer].status != LFAP_TIMER_RESET) {

          /* even if this timer is being used, doesn't mean there aren't any
             free timers.  check them all */

          #ifdef CS_DEBUG
          if (lfap_debug & LFAP_DBG_TIMERS) {
              fprintf(stderr, "current timer %u is being used ", current_timer);
              fprintf(stderr, " .. search for a different timer\n");
          }
          #endif

          for (timer = 0; timer < LFAP_MAX_TIMERS_OUTSTANDING; timer++) {
              if (lfap_timers[timer].status == LFAP_TIMER_RESET) {
                  current_timer = timer;
                  break;
              }
          }

          if (timer == LFAP_MAX_TIMERS_OUTSTANDING) {
              #ifdef CS_DEBUG 
              if (lfap_debug & LFAP_DBG_ERRORS)
                  ERR_errlog(0, lfap_efacid, LFAP_E_NO_TIMERS, 0);
              #endif
          }
      }

      #ifdef CS_DEBUG
      if (lfap_debug & LFAP_DBG_TIMERS) {
      fprintf(stderr, "SetSystemTimer: resetting timer: %u \n", current_timer);
      fprintf(stderr, "             setting for: %lu milliseconds\n", time_val);
      fprintf(stderr, "             setting for: %lu time ticks\n", time_ticks);
      }
      #endif

      lfap_timers[current_timer].status = LFAP_TIMER_WAIT;
      lfap_timers[current_timer].cb = (LFAP_CB_ENTRYPOINT) lfap_cb_entry;
      lfap_timers[current_timer].handle = return_handle;
      lfap_timers[current_timer].wait_time = time_val;

      status = NU_Reset_Timer(&lfap_timers[current_timer].timer, 
                             (expiration_routine)lfap_callback, 
                             time_ticks, 0, NU_ENABLE_TIMER);
      if (status != NU_SUCCESS) {
          #ifdef CS_DEBUG
          if (lfap_debug & LFAP_DBG_ERRORS)
              ERR_errlog(0, lfap_efacid, LFAP_E_TIMER_FAIL, 1, strerror(errno));
          #endif
      }
      else if (++current_timer == LFAP_MAX_TIMERS_OUTSTANDING) {
          current_timer = 0;
      }
  }
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPwriteError                                                  */
/*                                                                            */
/*  This function is called by the LFAP API to handle socket write errors.    */
/*                                                                            */
/******************************************************************************/
void LFAPHandleWriteError (int err)
{
  static int print_error = 0;
  unsigned long server;       /* temporary server address */
  struct teginfo tinfo;

  #ifdef CS_DEBUG
  if (lfap_debug & LFAP_DBG_ERRORS) {
    if (err != EAGAIN)
      ERR_errlog(0, lfap_efacid, LFAP_E_WRITEV_FAIL, 1, strerror(errno));
  }
  if (lfap_debug & LFAP_DBG_WARNINGS) {
    if ((err == EAGAIN) && (print_error%200 == 0))
      ERR_errlog(0, lfap_efacid, LFAP_W_WRITEV, 1, strerror(errno));
    print_error++;
  }
  #endif
  
  /* if error is broken pipe, we may not get the lost connection
     notification, because the socket event may never be triggered
     so we will never read the EOF, so, disconnect! */
  if ((err == EPIPE) && (lfap_get_status() == LFAP_STARTED)) {
    if ((server = LFAPGetActiveServer())) {
      /*                if (LFAPDisconnect(server) == LFAP_TCP_DISCONNECT_FAILED) */
      if (LFAPDisconnect(server) == LFAP_FAILURE)
        lfap_connection_lost(server);
      LFAPConnectionLost(server);
    }
    else if (lfap_debug & LFAP_DBG_ERRORS)
      ERR_errlog(0, lfap_efacid, LFAP_E_NO_DISCONNECT, 0);
  }
  
  /* The socket is full, set a flag so we don't keep trying */
  /* to write to it.                                        */
  if (err == EAGAIN)
  {
    lfap_socket_full = 1;
    /* set event for write */
    memset(&tinfo, 0, sizeof(tinfo));
    tinfo.ti_taskptr = lfapWorkerThread;
    tinfo.ti_tegbit = LFAP_EVENT_SOCKET_WRT_BIT;
    #ifndef LFAP_STANDALONE
    if (ioctl(fas_sd, SIOCSWRTEGBIT, (caddr_t) &tinfo) < 0) {
    #ifdef CS_DEBUG
      if (lfap_debug & LFAP_DBG_ERRORS)
        ERR_errlog(0, lfap_efacid, LFAP_E_IOCTL_FAIL, 1, strerror(errno));
    #endif
    }
    #endif            
  }
  
}


/******************************************************************************/
/*                                                                            */
/*  Function: LFAPSendMessage                                                 */
/*                                                                            */
/*  This function is called by the LFAP API to send the (void*) buf to the    */
/*  server.  buf_len includes the length of the header of the LFAP message.   */
/*  LFAP_SEND_FAILURE is returned if the send failed, LFAP_SEND_SUCCESS is    */
/*  returned if it was successful. Clear causes any partially sent messages   */
/*  to be cleared.                                                            */
/*                                                                            */
/******************************************************************************/
int LFAPSendMessage (void* buf, lfapui16_t buf_len, int clear)
{
    int wrote;
    static int len;
    static char *orig_buf;
    static char* ptr = 0;

    if (clear) {
      ptr = 0;
      return LFAP_SEND_SUCCESS;
    }
    

    if (ptr == 0) {
      ptr = buf;
      orig_buf = ptr;
      len = buf_len;
    }
    else {
      if (buf != orig_buf)
        fprintf(stderr, "LFAP-CODING_ERROR, switching buffers before send complete\n"); /* PAC_TEMPORARY - how to log */
    }
    
    
    while (len > 0) {
        if ((wrote = write(fas_sd, ptr, len)) <= 0) {
            LFAPHandleWriteError(errno);
            return LFAP_SEND_FAILURE;
        }

        len -= wrote;
        ptr += wrote;
    }

    ptr = 0;
    return LFAP_SEND_SUCCESS;
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPSendVector                                                  */
/*                                                                            */
/*  Note: writev changes the buffer pointer and length of each iovec          */
/*  structure it sends.  The buffer pointer is moved up by 'iov_len' bytes,   */
/*  and iov_len is set to 0 (or decremented by amount sent).                  */
/*                                                                            */
/******************************************************************************/
int LFAPSendVector (LFAPMsgHolder* vec, lfapui16_t vec_len,
                    lfapui16_t total_len,lfapui16_t lost_conn_count )
{
    int diff = 0;               /* bytes of a msg that was partially written */
    int wrote = 0;              /* bytes written in this pass */
    int vec_s = 0;              /* vector starting index */
    int vec_i = 0;              /* vector intermediate index */
    int written = 0;            /* temporary variable to count written bytes */
    static int len = 0;         /* length left to be written */
    static int vec_o = 0;       /* original vector length */
    static int total_wrote = 0; /* running total bytes written */
    static lfapui16_t prev_lost_conn_count = 0;
    static lfapui16_t partial_msg = 0;
    static struct iovec iov_t[LFAP_MAX_MSGS_IN_SEND_VECTOR];

    /* check to see if we didn't finish last time through */
    if (!vec_r) {
        total_wrote = 0;
        vec_o = vec_len;
        len = total_len;

        /* shuffle pointers into structure - in case don't write them all
           on first pass */
        for (vec_r = 0; vec_r < vec_len; vec_r++) {
            iov_t[vec_r].iov_base = vec[vec_r].buf;
            iov_t[vec_r].iov_len = vec[vec_r].len;
        }
    }
    else if (vec_len > vec_o) {
        /* check to see if more msgs have been added to the vector */

        len = total_len - total_wrote;
        for (; vec_o < vec_len; vec_o++) {
            iov_t[vec_r].iov_base = vec[vec_o].buf;
            iov_t[vec_r].iov_len = vec[vec_o].len;
            vec_r++;
        }
    }

    /* If we wrote part of a message last time and we have since lost */
    /* connection to the server, we DO NOT want to send the rest      */
    /* of the message to the new server.                              */
    /*                                                                */
    if (partial_msg && (prev_lost_conn_count != lost_conn_count))
    {
         int i;

         /* Even though we are skipping this first message we need */
         /* to count it as sent.                                   */
         len -= iov_t[0].iov_len;
         total_wrote += iov_t[0].iov_len;
         for (i = 1; i < vec_r; i++)
         {
              iov_t[i-1].iov_base = iov_t[i].iov_base;
              iov_t[i-1].iov_len = iov_t[i].iov_len;
         }
         vec_r--;         /* One less message in the send queue */
         partial_msg = 0; /* No longer have a partial message   */
    }
        
    while (len > 0) {
      if ((wrote = writev(fas_sd, iov_t, vec_r)) <= 0) {
        LFAPHandleWriteError(errno);
        return LFAP_SEND_FAILURE;
      }
      else {
        total_wrote += wrote;
        len -= wrote;
        partial_msg = 0; /* we wrote something so reset the flag */

        if (len > 0) {
          /* find message to begin with */
          written = 0;
          for (vec_i = 0; vec_i < vec_len; vec_i++) {
              written += vec[vec_i].len;
              if (written > total_wrote) {
                  /* bummers.  part of a message got sent, need to
                     adjust appropriately ... */
                  partial_msg = 1;
                  prev_lost_conn_count = lost_conn_count;
                  vec_s = vec_i + 1;
                  vec_r = vec_len - vec_i;
                  diff = written - total_wrote;
                  diff = vec[vec_i].len - diff;
                  iov_t[0].iov_base = (char *)vec[vec_i].buf + diff;
                  iov_t[0].iov_len = vec[vec_i].len - diff;
                  #ifdef CS_DEBUG
                  if (lfap_debug & LFAP_DBG_ERRORS) {
                    if (iov_t[0].iov_len < 0)
                      ERR_errlog(0, lfap_efacid, LFAP_E_INVALID_LENGTH, 0);
                  }
                  #endif
                  vec_i = 1;
                  break;
              }
              else if (written == total_wrote) {
                  /* whoo-hoo, landed on a boundry! */
                  vec_s = vec_i + 1;
                  vec_r = vec_len - vec_s;
                  vec_i = 0;
                  break;
              }
          }
          
          if ((vec_i == vec_len) || (len < 0)) {
              #ifdef CS_DEBUG
              if (lfap_debug & LFAP_DBG_ERRORS) {
               ERR_errlog(0,lfap_efacid,LFAP_E_WRITEV_FAIL,1,"length mismatch");
              }
              #endif
              vec_r = 0;
              return LFAP_SEND_FAILURE;
          }

          /* shuffle pointers for the rest of the messages */
          while (vec_s < vec_len) {
              iov_t[vec_i].iov_base = vec[vec_s].buf;
              iov_t[vec_i].iov_len = vec[vec_s].len;
              vec_i++; vec_s++;
          }

          #ifdef CS_DEBUG
          if (lfap_debug & LFAP_DBG_WARNINGS) {
          ERR_errlog(0,lfap_efacid,LFAP_W_WRITEV,1,"didn't finish in one pass");
          }
          #endif
        }
      }
    }

    /* reset vec_r, when function is called next time, will know that it is
       a new batch of messages to be sent */
    vec_r = 0;

    return LFAP_SEND_SUCCESS;
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPConnect                                                     */
/*                                                                            */
/*  Attempt to make a TCP connection to the server specified by server_ip.    */
/*                                                                            */
/******************************************************************************/
int LFAPConnect(lfapui32_t server_ip)
{
  /*    int lfap_connect = LFAP_TCP_CONNECTION_FAILED; */
    int lfap_connect = LFAP_FAILURE;
    struct sockaddr_in addr;
    struct teginfo tinfo;
    int nonblock;
    int rc;

    nonblock = 1;
    if ((fas_sd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
        #ifdef CS_DEBUG
        if (lfap_debug & LFAP_DBG_ERRORS)
            ERR_errlog(0, lfap_efacid, LFAP_E_SOCKET_FAIL, 1, strerror(errno));
        #endif
        return lfap_connect;
    }

    /* set event for read */
    memset(&tinfo, 0, sizeof(tinfo));
    tinfo.ti_taskptr = &LFAP_TCB;  /* NU_Current_Task_Pointer(); */
    tinfo.ti_tegbit = LFAP_EVENT_SOCKET_BIT;
    #ifndef LFAP_STANDALONE
    if (ioctl(fas_sd, SIOCSRDTEGBIT, (caddr_t) &tinfo) < 0) {
        #ifdef CS_DEBUG
        if (lfap_debug & LFAP_DBG_ERRORS)
            ERR_errlog(0, lfap_efacid, LFAP_E_IOCTL_FAIL, 1, strerror(errno));
        #endif
    }
    #endif

    /* set event for write */
    memset(&tinfo, 0, sizeof(tinfo));
    tinfo.ti_taskptr = lfapWorkerThread;
    tinfo.ti_tegbit = LFAP_EVENT_SOCKET_WRT_BIT;
    #ifndef LFAP_STANDALONE
    if (ioctl(fas_sd, SIOCSWRTEGBIT, (caddr_t) &tinfo) < 0) {
        #ifdef CS_DEBUG
        if (lfap_debug & LFAP_DBG_ERRORS)
          ERR_errlog(0, lfap_efacid, LFAP_E_IOCTL_FAIL, 1, strerror(errno));
     #endif
    }
    #endif


    memset(&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET; 
    addr.sin_port = INADDR_ANY; 
    addr.sin_addr.s_addr = htonl(server_ip); 
    addr.sin_port = htons(LFAP_PORT);

    ERR_errlog(0, lfap_efacid, LFAP_I_CONNECT, 2, inet_ntoa(addr.sin_addr), LFAP_PORT);

    if ((rc = connect(fas_sd, (struct sockaddr *)&addr, sizeof(addr))) == -1) {
        #ifdef CS_DEBUG
        if (lfap_debug & LFAP_DBG_ERRORS)
            ERR_errlog(0, lfap_efacid, LFAP_E_CONNECT_FAIL, 1, strerror(errno));
        #endif
        if (close(fas_sd) == -1) {
            #ifdef CS_DEBUG
            if (lfap_debug & LFAP_DBG_ERRORS)
                ERR_errlog(0, lfap_efacid, LFAP_E_CLOSE_FAIL, 1, strerror(errno));
            #endif
        }
        fas_sd = -1;
        return lfap_connect;
    }

    #ifndef LFAP_STANDALONE
    if (ioctl(fas_sd, FIONBIO, (caddr_t) &nonblock) < 0) {
        #ifdef CS_DEBUG
        if (lfap_debug & LFAP_DBG_ERRORS) {
            ERR_errlog(0, lfap_efacid, LFAP_E_IOCTL_FAIL, 1, strerror(errno));
        }
        #endif
    }
    #endif

    lfap_connection_established(server_ip);
    /*   return LFAP_TCP_CONNECTION_ESTABLISHED; */
    return LFAP_SUCCESS;
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPDisconnect                                                  */
/*                                                                            */
/*  Attempt to break the TCP connection to the server specified by server_ip. */
/*                                                                            */
/******************************************************************************/
int LFAPDisconnect(lfapui32_t server_ip)
{
  /*	int lfap_disconnect = LFAP_TCP_DISCONNECT_FAILED; */

  int lfap_disconnect = LFAP_FAILURE;
  
        if ((!server_ip) && (lfap_debug & LFAP_DBG_ERRORS))
            ERR_errlog(0, lfap_efacid, LFAP_E_DISCONNECT, 0);

        if (fas_sd != -1) {
            if (close(fas_sd) == -1) {
                #ifdef CS_DEBUG
                if (lfap_debug & LFAP_DBG_ERRORS)
                  ERR_errlog(0, lfap_efacid, LFAP_E_CLOSE_FAIL, 1, strerror(errno));
                #endif
            }
            else {
                fas_sd = -1;
                lfap_connection_lost(server_ip);
                /*                return LFAP_TCP_DISCONNECT_SUCCESS; */
                return LFAP_SUCCESS;
            }

            fas_sd = -1;
        }
        #ifdef CS_DEBUG
        else if (lfap_debug & LFAP_DBG_CONNECTING) {
            fprintf(stderr, "LFAPDisconnect FAILED, fas_sd is -1\n");
        }
        #endif
            
	return lfap_disconnect;
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPClearSocketFull                                             */
/*                                                                            */
/*  Clear the flag that indicates the LFAP socket is full.                    */
/*                                                                            */
/******************************************************************************/
void LFAPClearSocketFull()
{
  lfap_socket_full = 0;
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPGetSocketFull                                               */
/*                                                                            */
/*  If the LFAP socket is full and cannot be written to return 1 otherwise    */
/*  return 0                                                                  */
/*                                                                            */
/******************************************************************************/
int LFAPGetSocketFull()
{
  return lfap_socket_full;
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPLock                                                        */
/*                                                                            */
/*  See notes in header file                                                  */
/*                                                                            */
/******************************************************************************/
int LFAPLock()
{
  return NUE_Disable_Interrupts();
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPUnlock                                                      */
/*                                                                            */
/*  See notes in header file                                                  */
/*                                                                            */
/******************************************************************************/
void LFAPUnlock(int lockVal)
{
  NUE_Restore_Interrupts(lockVal);
}

#endif
