/*******************************************************************************
**
**    Cabletron Systems Incorporated
**    Post Office Box 5005
**    Rochester, NH 03866-5005
**    (c) Copyright Cabletron Systems Inc. 1999
**
**    Workfile:             %M%
**    Logfile:              %P%
**    Original Author:      Todd Crowley
**    SCCS Delta ID:        %I%
**    Last Delta Date/Time: %G% %U%
**
*******************************************************************************/

/*
 * message encoding routines
 */

/* Start of auto-generated includes */
#define INCLUDES_H_
#ifdef MLS_SOL_BUILD
#include "sol_includes.h"
#endif
#ifdef MLS_MIPS_BUILD
#include "mls_string.h"
#endif
/* End of auto-generated includes */

#include "Lfap.h"
#include "LfapPrivate.h"


/*
 * macros for encoding integers in network order
 */
#define ENCODE8(dst,src)  { *(dst)++ = (src)       & 0xff; }
#define ENCODE16(dst,src) { *(dst)++ = (src) >>  8 & 0xff; \
                            *(dst)++ = (src)       & 0xff; }
#define ENCODE32(dst,src) { *(dst)++ = (lfapui8_t)((src) >> 24 & 0xff); \
                            *(dst)++ = (lfapui8_t)((src) >> 16 & 0xff); \
                            *(dst)++ = (lfapui8_t)((src) >>  8 & 0xff); \
                            *(dst)++ = (lfapui8_t)((src)       & 0xff); }


/*
 *  _encode_ie_hdr
 *  Enoces the type/length header for an Information Element (IE)
 */
static int _encode_ie_hdr(LFAPBuffer *lbuf, lfapui16_t type, lfapui16_t len)
{
  lfapui8_t hdr[4], *ptr;

  if (!lbuf->form.beg) {

	/*
         * We're not inside a formatted record, so just append the IE header
         * to the end of the message.
         */

	if ((ptr = lfap_buffer_insert(lbuf, 4)) == NULL)
		return LFAP_ERROR;

	ENCODE16(ptr,type)
	ENCODE16(ptr,len)
  }
  else if (!lbuf->form.off) {

	/*
         * This is the first record, so we're creating the format on the fly.
         * We need to slide everything over, and add this header to the
         * record format field.
         */

	if (lfap_buffer_insert(lbuf, 4) == NULL)
		return LFAP_ERROR;

	ptr = lbuf->buf + lbuf->form.end;
	lfap_memmove(ptr + 4, ptr, lbuf->msglen - lbuf->form.end - 4);
	ENCODE16(ptr,type)
	ENCODE16(ptr,len)
	lbuf->form.end += 4;
  }
  else {

	/*
         * This is a formatted record, so all we have to do is ensure that
         * the header matches the format specifier.
         */

	/* encode the IE header in a temporary buffer */
	ptr = hdr;
	ENCODE16(ptr,type)
	ENCODE16(ptr,len)

	if (lbuf->form.off >= lbuf->form.end ||
	    lfap_memcmp(lbuf->buf + lbuf->form.off, hdr, 4)) {
		lbuf->err = LFAP_INVAL;
		return LFAP_ERROR;
	}

	lbuf->form.off += 4;
  }

  return LFAP_SUCCESS;
}


/*
 *  _encode_error
 *  Handles invalid type codes.
 */
static int _encode_error(LFAPBuffer *lbuf, LFAPIE *ie)
{
  lbuf->err = LFAP_INVAL;
  return LFAP_ERROR;
}


/*
 *  _encode_count
 *  Encode an LFAPDataCount IE
 */
static int _encode_count(LFAPBuffer *lbuf, LFAPIE *ie)
{
  lfapui8_t *ptr;

  if (_encode_ie_hdr(lbuf, ie->type, 16) ||
      (ptr = lfap_buffer_insert(lbuf, 16)) == NULL)
	return LFAP_ERROR;

  ENCODE32(ptr,ie->dataCount_ie.dc_rcvd.hi_val);
  ENCODE32(ptr,ie->dataCount_ie.dc_rcvd.lo_val);
  ENCODE32(ptr,ie->dataCount_ie.dc_sent.hi_val);
  ENCODE32(ptr,ie->dataCount_ie.dc_sent.lo_val);

  return LFAP_SUCCESS;
}


/*
 * _encode_flowid
 * Encode an LFAPFlowID IE
 */
static int _encode_flowid(LFAPBuffer *lbuf, LFAPIE *ie)
{
  lfapui8_t *ptr;
  lfapui16_t ie_len, true_len;
  lfapui32_t have_fas;

  /* Is the FAS portion blank? */
  have_fas = ie->fid_ie.fas_id[0] | ie->fid_ie.fas_id[1];

  /* ie_len is the 'special' length field encoded for a FlowID */
  ie_len  = have_fas ? 0x0804 : 0x0004;

  /* true_len is the actual length in bytes of the IE */
  true_len  = have_fas ? 12 : 4;

  if (_encode_ie_hdr(lbuf, ie->type, ie_len) ||
      (ptr = lfap_buffer_insert(lbuf, true_len)) == NULL)
        return LFAP_ERROR;

  if (have_fas) {
    ENCODE32(ptr,ie->fid_ie.fas_id[0]);
    ENCODE32(ptr,ie->fid_ie.fas_id[1]);
  }

  ENCODE32(ptr,ie->fid_ie.cce_id);

  return LFAP_SUCCESS;
}


/*
 *  _encode_octstr
 *  Encode an LFAPOctetStr IE
 */
static int _encode_octstr(LFAPBuffer *lbuf, LFAPIE *ie)
{
  lfapui8_t *ptr;
  lfapui16_t padlen;

  padlen = PADLEN(ie->octetStr_ie.len);

  if (_encode_ie_hdr(lbuf, ie->type, ie->octetStr_ie.len) ||
      (ptr = lfap_buffer_insert(lbuf, ie->octetStr_ie.len + padlen)) == NULL)
	return LFAP_ERROR;

  lfap_memcpy(ptr, ie->octetStr_ie.val, ie->octetStr_ie.len);
  lfap_memset(ptr + ie->octetStr_ie.len, 0, padlen);

  return LFAP_SUCCESS;
}


/*
 *  _encode_addr
 *  Encode an LFAPAddress IE
 */
static int _encode_addr(LFAPBuffer *lbuf, LFAPIE *ie)
{
  lfapui8_t *ptr;
  lfapui16_t rawlen, padlen;

  rawlen = 4 + ie->addr_ie.len;
  padlen = PADLEN(rawlen);

  if (_encode_ie_hdr(lbuf, ie->type, rawlen) ||
      (ptr = lfap_buffer_insert(lbuf, rawlen + padlen)) == NULL)
	return LFAP_ERROR;

  ENCODE16(ptr,ie->addr_ie.family);
  ENCODE16(ptr,ie->addr_ie.len);
  lfap_memcpy(ptr, ie->addr_ie.val, ie->addr_ie.len);
  lfap_memset(ptr + ie->addr_ie.len, 0, padlen);

  return LFAP_SUCCESS;
}


/*
 *  _encode_uint32
 *  Encode an LFAPUInt32 IE
 */
static int _encode_uint32(LFAPBuffer *lbuf, LFAPIE *ie)
{
  lfapui8_t *ptr;

  if (_encode_ie_hdr(lbuf, ie->type, 4) ||
      (ptr = lfap_buffer_insert(lbuf, 4)) == NULL)
	return LFAP_ERROR;

  ENCODE32(ptr,ie->i32_ie.val);

  return LFAP_SUCCESS;
}


/*
 *  _encode_uint16x2
 *  Encode an LFAPUInt16x2 IE
 */
static int _encode_uint16x2(LFAPBuffer *lbuf, LFAPIE *ie)
{
  lfapui8_t *ptr;

  if (_encode_ie_hdr(lbuf, ie->type, 4) ||
      (ptr = lfap_buffer_insert(lbuf, 4)) == NULL)
	return LFAP_ERROR;

  ENCODE16(ptr,ie->i16_ie.val[0]);
  ENCODE16(ptr,ie->i16_ie.val[1]);

  return LFAP_SUCCESS;
}

/*
 *  _encode_rsvd_uint16
 *  Encode a reserved followed by a uint16 IE
 */
static int _encode_rsvd_uint8(LFAPBuffer *lbuf, LFAPIE *ie)
{
  lfapui8_t *ptr;
  lfapui8_t *rptr;

  if (_encode_ie_hdr(lbuf, ie->type, 4) ||
      (ptr = lfap_buffer_insert(lbuf, 4)) == NULL)
	return LFAP_ERROR;

  rptr = (lfapui8_t *) &((LFAP_encode_decode_RsvdUint8 *)&ie->rsvd_ui8)->reserved_val;
  ENCODE8(ptr,*rptr++);
  ENCODE8(ptr,*rptr++);
  ENCODE8(ptr,*rptr);
  ENCODE8(ptr,ie->rsvd_ui8.val);

  return LFAP_SUCCESS;
}

/*
 *  _encode_rsvd_uint16
 *  Encode a reserved followed by a uint16 IE
 */
static int _encode_rsvd_uint16(LFAPBuffer *lbuf, LFAPIE *ie)
{
  lfapui8_t *ptr;

  if (_encode_ie_hdr(lbuf, ie->type, 4) ||
      (ptr = lfap_buffer_insert(lbuf, 4)) == NULL)
	return LFAP_ERROR;

  ENCODE16(ptr,ie->rsvd_ui16.reserved);
  ENCODE16(ptr,ie->rsvd_ui16.val);

  return LFAP_SUCCESS;
}

/*
 *  _encode_rsvd_uint24
 *  Encode a reserved followed by a uint24 IE
 */
static int _encode_rsvd_uint24(LFAPBuffer *lbuf, LFAPIE *ie)
{
  lfapui8_t *ptr;
  lfapui8_t *vptr;
  

  if (_encode_ie_hdr(lbuf, ie->type, 4) ||
      (ptr = lfap_buffer_insert(lbuf, 4)) == NULL)
	return LFAP_ERROR;

  vptr = (lfapui8_t *) &((LFAP_encode_decode_RsvdUint24 *)&ie->rsvd_ui24)->reserved_val;
  vptr++; /* skip over the reserved byte */
  ENCODE8(ptr,ie->rsvd_ui24.reserved);
  ENCODE8(ptr,*vptr++);
  ENCODE8(ptr,*vptr++);
  ENCODE8(ptr,*vptr);

  return LFAP_SUCCESS;
}

/*
 *  _encode_CoS
 *  Encode a Class of Service IE
 */
static int _encode_CoS(LFAPBuffer *lbuf, LFAPIE *ie)
{
  lfapui8_t *ptr;
  lfapui8_t tmp;

  if (_encode_ie_hdr(lbuf, ie->type, 4) ||
      (ptr = lfap_buffer_insert(lbuf, 4)) == NULL)
    return LFAP_ERROR;

  /* High order byte */
  tmp =  (lfapui8_t)ie->CoS.RT << 6;
  tmp |= (lfapui8_t)ie->CoS.CoS_type & 0x3F;
  ENCODE8(ptr, tmp);

  /* Middle high order byte */
  ENCODE8(ptr, ( (lfapui8_t) (ie->CoS.val >> 16) ));

  /* Middle low order byte */
  ENCODE8(ptr, ( (lfapui8_t) (ie->CoS.val >> 8) ));

  /* Low order byte */
  ENCODE8(ptr, ( (lfapui8_t) (ie->CoS.val) ));

  return LFAP_SUCCESS;
}


/*
 *  _encode_qualifier
 *  Encode an LFAPQualifierList IE
 */
static int _encode_qualifier(LFAPBuffer *lbuf, LFAPIE *ie)
{
  lfapui8_t *ptr;
  lfapui16_t len;
  int i;

  len = 0;
  for (i=0; i<ie->qual_ie.items; i++)
    len += 4 + ie->qual_ie.q_list[i].len + PADLEN(ie->qual_ie.q_list[i].len);

  if (_encode_ie_hdr(lbuf, ie->type, len) ||
      (ptr = lfap_buffer_insert(lbuf, len)) == NULL)
	return LFAP_ERROR;

  for (i=0; i<ie->qual_ie.items; i++) {
    ENCODE16(ptr,ie->qual_ie.q_list[i].id);
    ENCODE16(ptr,ie->qual_ie.q_list[i].len);
    lfap_memcpy(ptr, ie->qual_ie.q_list[i].val, ie->qual_ie.q_list[i].len);
    ptr += ie->qual_ie.q_list[i].len;
    len = PADLEN(ie->qual_ie.q_list[i].len);
	lfap_memset(ptr, 0, len);
	ptr += len;
  }

  return LFAP_SUCCESS;
}


/*
 *  _encode_prefix
 *  Encode an LFAPFlowIDPrefix IE
 */
static int _encode_prefix(LFAPBuffer *lbuf, LFAPIE *ie)
{
  lfapui8_t *ptr;

  if (_encode_ie_hdr(lbuf, ie->type, 8) ||
      (ptr = lfap_buffer_insert(lbuf, 8)) == NULL)
	return LFAP_ERROR;

  ENCODE32(ptr,ie->pfx.fas_id[0]);
  ENCODE32(ptr,ie->pfx.fas_id[1]);

  return LFAP_SUCCESS;
}


/* prototype of the '_encode' functions */
typedef int (*vptr)(LFAPBuffer*, LFAPIE*);

vptr _encode_vtbl[LFAP_MAX_IE_CODE];

void lfap_init_encode() {
  int ie;

  /* Initialize array to all errors */
  for(ie = 0; ie < LFAP_MAX_IE_CODE; ie++) {
    _encode_vtbl[ie] = _encode_error;
  }

  /* Fill in the known IE encode routines */
  _encode_vtbl[LFAP_TRANSPORT_FAS_IP]          = _encode_addr;
  _encode_vtbl[LFAP_TRANSPORT_MULTIPLE_RECORD] = _encode_error;
  _encode_vtbl[LFAP_FLOW_ID]                   = _encode_flowid;
  _encode_vtbl[LFAP_SRC_ADDR]                  = _encode_addr;
  _encode_vtbl[LFAP_DST_ADDR]                  = _encode_addr;
  _encode_vtbl[LFAP_FLOW_QUALIFIER]            = _encode_qualifier;
  _encode_vtbl[LFAP_TIME]                      = _encode_uint32;
  _encode_vtbl[LFAP_UTC_TIME]                  = _encode_uint32;
  _encode_vtbl[LFAP_DELTA_TIME]                = _encode_error; /* Not yet implemented */
  _encode_vtbl[LFAP_CLASS_OF_SERVICE]          = _encode_CoS;
  _encode_vtbl[LFAP_SRC_CCE_ADDR]              = _encode_addr;
  _encode_vtbl[LFAP_CLIENT_DATA]               = _encode_octstr;
  _encode_vtbl[LFAP_COMMAND_CODE]              = _encode_rsvd_uint8;
  _encode_vtbl[LFAP_IE_LIST]                   = _encode_error; /* Not yet implemented */
  _encode_vtbl[LFAP_FAS_IP_ADDR]               = _encode_addr;
  _encode_vtbl[LFAP_FAIL_CODE]                 = _encode_rsvd_uint8;
  _encode_vtbl[LFAP_FLOW_ID_PREFIX]            = _encode_prefix;
  _encode_vtbl[LFAP_FLOW_STATE]                = _encode_rsvd_uint8;
  _encode_vtbl[LFAP_BYTES_TOTAL]               = _encode_count;
  _encode_vtbl[LFAP_BYTES_DELTA]               = _encode_count;
  _encode_vtbl[LFAP_PACKETS_TOTAL]             = _encode_count;
  _encode_vtbl[LFAP_PACKETS_DELTA]             = _encode_count;
  _encode_vtbl[LFAP_PROTOCOL_ID]               = _encode_octstr;
  _encode_vtbl[LFAP_SRC_PORT_UDP]              = _encode_uint32;
  _encode_vtbl[LFAP_SRC_PORT_TCP]              = _encode_uint32;
  _encode_vtbl[LFAP_SRC_SOCKET_IPX]            = _encode_uint32;
  _encode_vtbl[LFAP_SRC_AS]                    = _encode_uint32;
  _encode_vtbl[LFAP_DEST_AS]                   = _encode_uint32;
  _encode_vtbl[LFAP_INGRESS_PORT]              = _encode_uint32;
  _encode_vtbl[LFAP_EGRESS_PORT]               = _encode_uint32;
  _encode_vtbl[LFAP_VLAN_ID]                   = _encode_uint16x2;
  _encode_vtbl[LFAP_MPLS_IN_LABEL]             = _encode_uint32;
  _encode_vtbl[LFAP_MPLS_OUT_LABEL]            = _encode_uint32;
  _encode_vtbl[LFAP_NEXT_HOP_ADDR]             = _encode_addr;
  _encode_vtbl[LFAP_LDP_FEC]                   = _encode_octstr; /* treated as octet string on output side */
  /* The rest of the IE's are not yet implemented */
}

/*
 *  lfap_add_IE
 *  Encodes a generic LFAPIE into an LFAPBuffer
 */
int lfap_add_IE(LFAPBuffer *lbuf, LFAPIE *ie)
{
  return CALL_VIRTUAL(ie,_encode)(lbuf,ie);
}


/*
 *  lfap_begin_message
 *  Encodes the message header of an LFAP message.
 */
int lfap_begin_message(LFAPBuffer *lbuf, LFAPHeader *hdr)
{
  lfapui8_t *ptr;

  if (lfap_buffer_get_msglen(lbuf) != 0) {
	lbuf->err = LFAP_NOTEMPTY;
	return LFAP_ERROR;
  }

  if ((ptr = lfap_buffer_insert(lbuf, LFAP_HEADER_SIZE)) == NULL)
	return LFAP_ERROR;

  ENCODE8(ptr,LFAP_VERSION);
  ENCODE8(ptr,hdr->opcode);
  ENCODE8(ptr,0);
  ENCODE8(ptr,hdr->status);
  ENCODE16(ptr,hdr->msgid);
  ENCODE16(ptr,0);

  return LFAP_SUCCESS;
}


/*
 *  lfap_begin_multiple
 *  Marks the start of a multiple-record IE
 */
int lfap_begin_multiple(LFAPBuffer *lbuf)
{
  lfapui8_t *ptr;

  if ((ptr = lfap_buffer_insert(lbuf, 8)) == NULL)
	return LFAP_ERROR;

  lbuf->mlt.beg = ptr - lbuf->buf;

  ENCODE16(ptr,LFAP_TRANSPORT_MULTIPLE_RECORD);
  lfap_memset(ptr, 0, 6);

  return LFAP_SUCCESS;
}


/*
 *  lfap_begin_fixed
 *  Marks the start of the fixed data portion of a multiple-record IE
 */
int lfap_begin_fixed(LFAPBuffer *lbuf)
{
  return LFAP_SUCCESS;
}


/*
 *  lfap_end_fixed
 *  Marks the end of the fixed data portion of a multiple-record IE
 */
int lfap_end_fixed(LFAPBuffer *lbuf)
{
  lfapui8_t *ptr;
  lfapui16_t fixedlen;

  ptr = lbuf->buf + lbuf->mlt.beg + 4;
  fixedlen = lbuf->msglen - lbuf->mlt.beg - 8;
  ENCODE16(ptr, fixedlen);

  return LFAP_SUCCESS;
}


/*
 *  lfap_begin_record
 *  Marks the start of a data record that should be added to the current
 *  mutliple-record IE.
 */
int lfap_begin_record(LFAPBuffer *lbuf)
{
  if (!lbuf->form.beg)
	lbuf->form.beg = lbuf->form.end = lbuf->msglen;

  return LFAP_SUCCESS;
}


/*
 *  lfap_begin_record
 *  Marks the end of the current data record.
 */
int lfap_end_record(LFAPBuffer *lbuf)
{
  lfapui8_t *ptr;
  lfapui16_t formlen;

  if (!lbuf->form.off) {
	formlen = lbuf->form.end - lbuf->form.beg;
	ptr = lbuf->buf + lbuf->mlt.beg + 6;
	ENCODE16(ptr,formlen);
  }
  else if (lbuf->form.off != lbuf->form.end) {
	lbuf->err = LFAP_INVAL;
	return LFAP_ERROR;
  }

  lbuf->form.off = lbuf->form.beg;

  return LFAP_SUCCESS;
}


/*
 *  lfap_end_multiple
 *  Marks the end of the current multiple-record IE
 */
int lfap_end_multiple(LFAPBuffer *lbuf)
{
  lfapui8_t *ptr;
  lfapui16_t ie_len;

  ie_len = lbuf->msglen - lbuf->mlt.beg - 4;

  ptr = lbuf->buf + lbuf->mlt.beg + 2;
  ENCODE16(ptr,ie_len);

  lbuf->mlt.beg = 0;
  lbuf->form.beg = 0;
  lbuf->form.off = 0;

  return LFAP_SUCCESS;
}


/*
 *  lfap_end_message
 *  Marks the end of the message being encoded.
 */
int lfap_end_message(LFAPBuffer *lbuf)
{
  lfapui8_t *ptr;
  lfapui16_t msglen;

  msglen = lbuf->msglen - LFAP_HEADER_SIZE;
  ptr = lbuf->buf + 6;
  ENCODE16(ptr,msglen);

  return LFAP_SUCCESS;
}
