/* 
**    Riverstone Networks, Inc
**    5200 Great America Parkway
**    Santa Clara CA 95054
**    (c) Copyright Riverstone Networks 2000,2001


**  Trivial Flow Accounting Server for use with
**  Switching gear supporting LFAPv4 protocol
** 
**  Authors:
**  Paul Calato calato@riverstonenet.com
**  Mike MacFaden mrm@riverstonenet.com 
* 
*/


#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <time.h>


#ifndef _WIN32
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>


/* remap some items used by windows */
typedef int SOCKET;
#define closesocket close


#ifndef HAVE_LONGLONG_T
typedef long long longlong_t;
#endif
#else


/* win32 definsions */
#define longlong_t __int64
#endif



#include "Lfap.h"
#include "LfapMsg.h"
#include "LfapVR.h"
#include "LfapVRA.h"
#include "LfapCR.h"
#include "LfapCAN.h"
#include "LfapCRN.h"
#include "LfapFER.h"
#include "LfapFAR.h"
#include "LfapFUN.h"
#include "LfapAR.h"
#include "LfapARA.h"
#include "LfapKA.h"
#include "LfapDR.h"
#include "LfapDebug.h"

#define LFAP_BUFFER_SIZE 4096


int sfas_debug = 0;                 /* debug flag */
int loose_handshake = 1;            /* Strict protocol support off by default */
enum { keep_alive_interval = 20 };  /* seconds, matches to config command: 
                                      lfap set lost-contact-interval 20..2000 seconds
                                      application acks needed in case tcp session is not 
                                      dropped by server side dies anyway
                                      RS/SSR have a default of 60 seconds in which it must 
                                      receive some msg from fas.
                                    */

time_t ar_time;     /* the time when the AR GET_TIME message was sent */
time_t ara_time;    /* the time when the ARA resonse came back        */
enum {MAX_AR_ARA_TIME_WINDOW = 10}; /* max time between AR and ARA response   */
int ar_time_msg_id = 0 ;    /* The AR message ID, so we can match the ARA     */
unsigned long  switch_time; /* switch uptime at the time the ARA was sent     */
enum { SWITCH_TICS_PER_SEC = 100 };

typedef struct sfas_stat  {
  int bad_opcode;
  int session_establishment_error;
  int corrupted_msg;
  int version_mismatch;
  int unexpected_ara;
  int unexpected_vra;    
  int bad_read;
  int bad_version_rx;
  int lost_session;
  int connection_count;
  int msgs;
  int bytes;
  int keepalives_sent;
    
  int vr;
  int cr;
  int farc;
  int fun;
  int vra;
  int ar;
  int ara;
  
} SFAS_STATS_T;


enum {

  /* WARNING - the order here is the order of output */

  /* COMMON */
  FLOW_ID,
  TIME,

  /* FAR RECORDS */
  SRC_ADDR,
  DEST_ADDR,
  IS_MPLS,
  SRC_PORT,
  DEST_PORT,
  PROTOCOL,
  COS_ARR, 
  INGRESS_PORT,
  EGRESS_PORT,
  SRC_AS,
  DEST_AS,
  SRC_CCE_ADDR,
  CHECKPOINT,
  CXN_PRIORITY,
  IN_MPLS_LABEL,
  OUT_MPLS_LABEL,
  LDP_FEC,
  NEXT_HOP_ADDR,
  SRC_VLAN_ID,
  DEST_VLAN_ID,

  /* FUN RECORDS */
  FUN_STATE,
  BYTES,
  PKTS,

  MAX_PARAMS,

  IP_FLAG   = 0,
  MPLS_FLAG = 1,
  FUN_FLAG  = 2

};

#define MAX_PARAM_SIZE 100
char output_params[MAX_PARAMS][MAX_PARAM_SIZE];
unsigned char output_flag[MAX_PARAMS][3];
char *output_strings[MAX_PARAMS];

static SFAS_STATS_T stats;


#define SHOW_DEBUG_MSG(string)  if (sfas_debug) printf("%s(%d): %s\n",__FILE__, __LINE__, string)

#define SHOW_DEBUG_MSG1(string, x) \
   { if (sfas_debug) { \
     printf(string, x); \
     printf("\n"); \
   }  }


#define SHOW_DEBUG_MSG2(string, x, y)\
        { if(sfas_debug) { \
        printf(string,x,y); \
        printf("\n"); \
    } }


#define SHOW_DEBUG_MSG3(string, x, y, z)\
        { if(sfas_debug) { \
        printf(string,x,y,z); \
        printf("\n"); \
    } }


#define SHOW_DEBUG_MSG4(string, x, y, z, a)\
        { if(sfas_debug) { \
        printf(string,x,y,z,a); \
        printf("\n"); \
    } }



#define SHOW_DEBUG_MSG5(string, x, y, z, a, b)\
        { if(sfas_debug) { \
        printf(string,x,y,z,a,b); \
        printf("\n"); \
    }}



/* this program's version info */
const short maj_version     = 5; /* corresponds to LFAP version */
const short min_version     = 0; /* functional enahcements      */
const short update_version  = 1; /* bug fixes only              */


/* The following states represent the session  
 * establishment sequence for a CCE to send flow  
 * records to a FAS
 */
typedef enum fas_state {
  INIT=1,
  NOT_CONNECTED,
  CONNECT,
  PROTOCOL_NEGOTIATION,
  CONNECTED,
  DATA_NEGOTIATION,
  SEND
} fas_state_t;


fas_state_t the_fas_state = INIT;


#define MAX_LISTEN_QUEUED (1)
#define MAX_IE_IN_FUN (10)


/* LFAPv5 Strings */

char *op_strings[LFAP_MAX_OPCODE+1]; 
char *op_cmd_codes[LFAP_MAX_CMD_CODE+1]; 
char *ie_codes[LFAP_MAX_IE_CODE+1];

char *protocol_list[] = {
    "reserved",
    "icmp",
    "igmp",
    "ggp",
    "ip",
    "st",
    "tcp",
    "ucl",
    "egp",
    "igp",
    "bbn-rrc-mon",
    "nvp-II",
    "pup",
    "argus",
    "emcon",
    "xnet",
    "chaos",
    "udp",
    "mux",
    "dcn-meas",
    "hmp",
    "prm",
    "xns-idp",
    "trunk-1",
    "trunk-2",
    "leaf-1",
    "leaf-2",
    "drp",
    "irtp",
    "iso-tp4",
    "netblt",
    "mfe-nsp",
    "merit-inp",
    "sep",
    "3pc",
    "idpr",
    "xtp",
    "ddp",
    "idpr-cmtp",
    "tp++",
    "il",
    "sip",
    "sdrp",
    "sip-sr",
    "sip-frag",
    "idrp",
    "rsvp",
    "gre",
    "mhrp",
    "bna",
    "sipp-esp",
    "sipp-ah",
    "I-nlsp",
    "swipe",
    "nhrp",
    "",
    "",
    "",
    "",
    "",
    "",
    "host internal protocl",
    "cftp",
    "sat-expak",
    "kryptolan",
    "rvd",
    "ippc",
    "any dist file system",
    "sat-mon",
    "visa",
    "ipcv",
    "cpnx",
    "cphb",
    "wsn",
    "pvp",
    "",
    "br-sat-mon",
    "sun-nd",
    "wb-mon",
    "wb-expak",
    "iso-ip",
    "vmtp",
    "secure-vmtp",
    "vines",
    "ttp",
    "nsfnet-igp",
    "dgp",
    "tcf",
    "igrp",
    "ospfigp",
    "sprite-rpc",
    "larp",
    "mtp",
    "ax.25",
    "ipip",
    "micp",
    "scc-sp",
    "etherip",
    "encap",
    "any private encryption",
    "gmtp",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
    "",
"reserved"
};


static int connection_count = 0;

void
dump_buffer(char *pb, int len)
{
    int count;
    
    printf("dump: len = %d\n", len);
    
    for (count = 0; count < len; count++) {
        printf("%02X ", pb[count]);
        if ((count % 16) == 15)
            printf("\n");
    }
    printf("\n\n");
}


/* pre: sd is a connected tcp socket, buf is valid buffer of len size
 * post: buf contains len bytes of data read from sd
 *       returns 0 on success else -1
 *       NOTES: blocks in read till all len bytes are read
 */
int
tcp_readn(SOCKET sd, char *buf, int len)
{
    int rc;
    int total = len;


    char *ptr = buf;
 
    while (len > 0) {
        rc = recv(sd, ptr, len, 0);
        if (rc == 0)
            return 0;
        else if (rc == -1)
            return -1;
 
        ptr += rc;
        len -= rc;
    }
    return total;
}

/*
 * pre: Connect state or higher
 * post: if error on send move to Not Connected state
 */
void
send_it(SOCKET client, unsigned char *buf, unsigned int len)
{
    int rc;


    rc = send(client, (const char *)buf, len, 0);
    if (rc == -1) {
        perror("send: ");
        the_fas_state = NOT_CONNECTED;
    }
}


/* pre: LFAP_PORT is not in use on OS
 * post: return socket in listen state
 */
int
create_server()
{
    struct sockaddr_in sa;
    int opt = 1;
    
    SOCKET sck = socket(PF_INET, SOCK_STREAM, 0);
    if (sck == -1) {
                perror("socket");
        return -1;
        }
    if (setsockopt(sck, SOL_SOCKET, SO_REUSEADDR,
                   (char *)&opt, sizeof opt) == -1) {
        perror("setsockopt");
        closesocket(sck);
        return -1;
    }
    
    memset(&sa, 0, sizeof sa);
    sa.sin_family = AF_INET;
    sa.sin_port = htons(LFAP_PORT);
    sa.sin_addr.s_addr = INADDR_ANY;
    
    if (bind(sck, (struct sockaddr *)&sa, sizeof(sa)) == -1) {
        perror("bind");
        closesocket(sck);        
        return -1;
    }


    if (listen(sck, MAX_LISTEN_QUEUED) == -1 ) {
        perror("listen");
        closesocket(sck);        
        return -1;
    }
    
    return sck;
}


/* Used to keep CCE from timing out connection 
 */
void
send_keepalive(SOCKET client)
{
    LFAP_KA  ka;
    char ka_buffer[LFAP_BUFFER_SIZE];


    initLfapKA (&ka);    
    lfap_buffer_init( &ka.lfap_msg, ka_buffer, sizeof(ka_buffer));
    encodeCompletedLfapKA(&ka);

    send_it(client, ka.lfap_msg.buf, ka.lfap_msg.msglen);
    
    SHOW_DEBUG_MSG("send_ka: wrote keepalive ka %d bytes");
    stats.keepalives_sent++;
}

void
send_ar_get_time(SOCKET client)
{
  LFAP_AR ar;
  char ar_buffer[LFAP_BUFFER_SIZE];
  
  /* request time so we can calcuate wall clock time */
  initLfapAR( &ar );
  lfap_buffer_init( &ar.lfap_msg, ar_buffer, sizeof(ar_buffer));
  setCommandLfapAR(&ar, LFAP_RETURN_TIME);
  encodeCompletedLfapAR( &ar );
  ar_time_msg_id = getLfapMsgId(&ar);
  send_it(client, ar.lfap_msg.buf, ar.lfap_msg.msglen);
  time(&ar_time);
}

void
send_fer(SOCKET client)
{
    LFAP_FER  fer;
    char fer_buffer[LFAP_BUFFER_SIZE];

    initLfapFER (&fer);    
    lfap_buffer_init( &fer.lfap_msg, fer_buffer, sizeof(fer_buffer));
    encodeCompletedLfapFER(&fer);

    send_it(client, fer.lfap_msg.buf, fer.lfap_msg.msglen);
    if (the_fas_state != NOT_CONNECTED)
      the_fas_state = SEND;
    
    printf("sent FER\n");
}

/* pre: Connect state or Protocol Negotiation
 * post: send VRA
 *       Move to connected state or  Protocol Negotiation state
 */
void
process_vr(SOCKET client, char *buffer, size_t len)
{
    LFAP_VR vr;
    LFAP_VRA vra;
    char vra_buffer[LFAP_BUFFER_SIZE];


    stats.vr++;
    initRcvdLfapVR( &vr, buffer, len );
    initLfapVRAfromVR( &vra, &vr );


    if (the_fas_state != CONNECT && the_fas_state != PROTOCOL_NEGOTIATION) {
      printf("Got VR out of sequence\n");
      the_fas_state = NOT_CONNECTED;
      stats.session_establishment_error++;
      return;
    }



    if (verifyLfapVR(&vr) == LFAP_VR_INVALID) {
      setLfapMsgStatus(&vra,LFAP_MSG_CORRUPTED);
      printf("Got bad VR msg\n");
      the_fas_state = NOT_CONNECTED;
      stats.corrupted_msg++;
    }
    else {
      the_fas_state = PROTOCOL_NEGOTIATION;
      if ( lfap_msg_get_version(buffer) != LFAP_VERSION ) {
        /* The spec says we need to send back the highest supported
         * version in the VRA message. Only one is support here and
         * our version is already in the message header.
         */
        setLfapMsgStatus(&vra, LFAP_MSG_VERSION);
        ++stats.version_mismatch;
      }
      else {
        the_fas_state = CONNECTED;
      }
    }
    
    printf("VR received\n");
    lfap_buffer_init(&vra.lfap_msg, vra_buffer, sizeof(vra_buffer));
    encodeCompletedLfapVRA( &vra );
    send_it(client, vra.lfap_msg.buf, vra.lfap_msg.msglen);
}



/* pre: received VR message and sent VRA
 *      Connected State 
 * post: Send connection accpeted notification (CAN)
 *       Move to Data Negotiation state
 */
void
process_cr(SOCKET client, char *buffer, int len)
{
    LFAP_CR cr;
    LFAP_CAN can;
    char can_buffer[LFAP_BUFFER_SIZE];
    struct in_addr IPaddr;

    stats.cr++;
    initRcvdLfapCR( &cr, buffer, len );
    if ( verifyLfapCR(&cr) == LFAP_CR_INVALID) {
       printf("Got bad CR msg\n");
       the_fas_state = NOT_CONNECTED;
       stats.corrupted_msg++;
       return;
    }

    SHOW_DEBUG_MSG("Processing CR");
    if (the_fas_state != CONNECTED) {
      printf("Got CR message out of handshake sequence\n");
      the_fas_state = NOT_CONNECTED;
      stats.session_establishment_error++;
      return;
    }

      
    /* FAS IP Address processing :
       Since the verify ensures that mandatory IE's
       are present, we don't need to check for FAS IP IE
    */
    printf("CR received\n");
    do {
      IPaddr = *(struct in_addr*)&cr.fas_ip_addr.addr_ie.val[0];
      printf(" FAS Address: %s\n", inet_ntoa(IPaddr));
    } while ( decodeNextRecordLfapCR( &cr ) != LFAP_EMPTY_BUFFER );
    
    /* we can now send the CONNECTION ACCEPTED NOTIFICATION */
    initLfapCAN( &can );
    lfap_buffer_init( &can.lfap_msg, can_buffer, sizeof(can_buffer));
    encodeCompletedLfapCAN( &can );
    send_it(client, can.lfap_msg.buf, can.lfap_msg.msglen);
    
    /* If we successfully sent the message, we can move to
       the next state.
    */
    if (the_fas_state == CONNECTED)
      the_fas_state = DATA_NEGOTIATION ;
}

/* pre: Data Negotiation State 
 * post: No change
 */
void
process_ar(SOCKET client, char *buffer, int len)
{
    LFAP_AR ar;
    LFAP_ARA ara;
    char ara_buffer[LFAP_BUFFER_SIZE];
    unsigned int verify;
    unsigned int pfx[2] = {1,2};
    struct in_addr IPaddr;

    stats.ar++;
    initRcvdLfapAR( &ar, buffer, len );
    initLfapARAfromAR( &ara, &ar );

    if (the_fas_state < DATA_NEGOTIATION ) {
      printf("Got AR message out of handshake sequence\n");
      the_fas_state = NOT_CONNECTED;
      stats.session_establishment_error++;
      return;
    }

    if ( (verify = verifyLfapAR(&ar)) == LFAP_AR_INVALID) {
       setLfapMsgStatus(&ara,LFAP_MSG_CORRUPTED);
       printf("Got bad AR msg\n");
       the_fas_state = NOT_CONNECTED;
       return;
    }

    SHOW_DEBUG_MSG1("Processing command %s", op_cmd_codes[ar.command.rsvd_ui8.val]);
    switch ( ar.command.rsvd_ui8.val ) {
      
    case LFAP_RETURN_FLOW_PREFIX:
      printf("AR - flow prefix\n");
      setFlowPrefixLfapARA( &ara, pfx[0], pfx[1]);
      break;
      
    case LFAP_LIST_OF_FASS:
      /* List of FAS processing here */
      printf("AR - list FASs\n");
      do {
        if ( verify & LFAP_AR_FAS_IP_ADDR) {
          IPaddr = *(struct in_addr*)&ar.fas_ip_addr.addr_ie.val[0];
          printf(" FAS Address: %s\n", inet_ntoa(IPaddr));
        }
      } while ( decodeNextRecordLfapAR( &ar ) != LFAP_EMPTY_BUFFER );
      break;
      
    default:
      printf( "Unexpected AR Command");
      break;
    }

    /* Send the acknowledge back */
    lfap_buffer_init( &ara.lfap_msg, ara_buffer, sizeof(ara_buffer));
    encodeCompletedLfapARA( &ara );
    send_it( client, ara.lfap_msg.buf, ara.lfap_msg.msglen);
}


/* pre: DATA_NEGOTIATION state
        sent AR GET_TIME
 * post: store time value
 *       Move to SEND state
 */
void
process_ara(SOCKET client, char *buffer, int len)
{
    LFAP_ARA ara;
    int time_diff;

    /* When we got the ARA should be very close to when
       the CCE sent it so we'll use this wall clock time
       to indicate when flows started or ended
    */
    time(&ara_time);

    stats.ara++;
    initRcvdLfapARA( &ara, buffer, len );
    if ( verifyLfapARA(&ara) == LFAP_ARA_INVALID) {
       printf("Got bad ARA msg\n");
       the_fas_state = NOT_CONNECTED;
       stats.corrupted_msg++;
       return;
    }

    SHOW_DEBUG_MSG("Processing ARA");
    if (the_fas_state != DATA_NEGOTIATION) {
      printf("Got ARA message out of handshake sequence\n");
      the_fas_state = NOT_CONNECTED;
      stats.session_establishment_error++;
      return;
    }

      
    /* ARA time value processing :
     * Since the verify ensures that mandatory IE's
     * are present, we don't need to check for time value being present
    */
    printf("ARA received\n");
    if ( ar_time_msg_id != getLfapMsgId(&ara) ) {
      printf("ARA response message ID did not equal that of the AR request message ID\n");
    }
    else {
      time_diff = ara_time - ar_time;
      if (time_diff > MAX_AR_ARA_TIME_WINDOW) {
        printf("AR - GET_TIME request did not get a response in %d seconds\n", MAX_AR_ARA_TIME_WINDOW);
        printf("Trying again...\n");
        send_ar_get_time(client);
      }
      else {
        switch_time = ara.time.i32_ie.val;
        printf("Time Mark = %s", ctime(&ara_time));
        send_fer(client);
      }
    }
}


char *
pro2str(int protocol)
{
    if (protocol > 0 && protocol < sizeof(protocol_list)/ sizeof(char *))
        return protocol_list[protocol];

    return "n/a";
}



/* output to stdout a far record
 */
void
get_common_output_params(LFAP_FAR *lfap_far, int verify)
{
    int  k;
    struct in_addr IPaddr;
    const char *NOT_IP = "(not IP address)";
    
    /* Flow ID */
    if ( verify & LFAP_FAR_FLOW_ID) {
      sprintf(&output_params[FLOW_ID][0], "F:%lu.%lu.%lu",
             lfap_far->flow_id.fid_ie.fas_id[0], 
             lfap_far->flow_id.fid_ie.fas_id[1], 
             lfap_far->flow_id.fid_ie.cce_id);
    }
    else
      sprintf(&output_params[FLOW_ID][0], "F:Error");

    /* Start time */
    if (verify & LFAP_FAR_TIME) {
      char *ti;
      time_t t;
      t = ara_time + ((lfap_far->time.i32_ie.val - switch_time) / SWITCH_TICS_PER_SEC );
      ti = ctime(&t);
      ti[strlen(ti) - 1] = 0;
      sprintf(&output_params[TIME][0], "%s", ti);
    }
    else
      sprintf(&output_params[TIME][0], "Error");


    /* Src Address */
    if (verify & LFAP_FAR_SRC_ADDR) {
      IPaddr =  *(struct in_addr*)&lfap_far->src_addr.addr_ie.val[0];
      sprintf(&output_params[SRC_ADDR][0], "%s", lfap_far->src_addr.addr_ie.family == 1 ? inet_ntoa(IPaddr) : NOT_IP);
    }


    /* Dest Address */
    if ( verify & LFAP_FAR_DEST_ADDR) {
      IPaddr = *(struct in_addr*)&lfap_far->dest_addr.addr_ie.val[0];
      sprintf(&output_params[DEST_ADDR][0], "%s", lfap_far->dest_addr.addr_ie.family == 1 ? inet_ntoa(IPaddr) : NOT_IP);
    }


    /* Ingress ifIndex */
    if ( verify & LFAP_FAR_INGRESS_PORT )
      sprintf(&output_params[INGRESS_PORT][0], "%lu", lfap_far->ingress_port.i32_ie.val);

    /* Egress ifIndex */
    if ( verify & LFAP_FAR_EGRESS_PORT )
      sprintf(&output_params[EGRESS_PORT][0], "%lu", lfap_far->egress_port.i32_ie.val);


    /* Src CCE address */
    if (verify & LFAP_FAR_SOURCE_CCE_ADDR) {
      IPaddr = *(struct in_addr *)&lfap_far->src_cce_addr.addr_ie.val[0];
      sprintf(&output_params[SRC_CCE_ADDR][0], "%s", lfap_far->src_cce_addr.addr_ie.family == 1 ? inet_ntoa(IPaddr) : NOT_IP);
    }

    if (verify & LFAP_FAR_QUALIFIER) {
      /* Checkpoint (5-minute, 15-minute, etc...) */
        for( k=0; k < lfap_far->qualifier.qual_ie.items; k++ ) {
          if(lfap_far->qualifier.qual_ie.q_list[k].id == LFAP_CHECKPOINT_ID) {
          switch( ntohl(*((unsigned long*)lfap_far->qualifier.qual_ie.q_list[k].val) ) ) {
            case LFAP_CHECKPOINT_HOURLY:
              sprintf(&output_params[CHECKPOINT][0], "hourly" );
              break;
            case LFAP_CHECKPOINT_DAILY:
              sprintf(&output_params[CHECKPOINT][0], "daily" );
              break;
            case LFAP_CHECKPOINT_WEEKLY:
              sprintf(&output_params[CHECKPOINT][0], "weekly" );
              break;
            case LFAP_CHECKPOINT_5_MINUTE:
              sprintf(&output_params[CHECKPOINT][0], "5-minute" );
              break;
            case LFAP_CHECKPOINT_15_MINUTE:
              sprintf(&output_params[CHECKPOINT][0], "15-minute" );
              break;
            default:
              sprintf(&output_params[CHECKPOINT][0], "Error" );
              break;
          }
        }
        else if( lfap_far->qualifier.qual_ie.q_list[k].id == LFAP_CXN_PRIORITY_ID ) {
          /* Priority (low, high, etc...)  */
          switch( htonl(*((unsigned long*)lfap_far->qualifier.qual_ie.q_list[k].val )) ) {
            case LFAP_CXN_PRIORITY_PREMIUM:
              sprintf(&output_params[CXN_PRIORITY][0], "premium" );
              break;
            case LFAP_CXN_PRIORITY_HIGH:
              sprintf(&output_params[CXN_PRIORITY][0], "high" );
              break;
            case LFAP_CXN_PRIORITY_MEDIUM:
              sprintf(&output_params[CXN_PRIORITY][0], "med" );
              break;
            case LFAP_CXN_PRIORITY_LOW:
              sprintf(&output_params[CXN_PRIORITY][0], "low" );
              break;
            default:
              sprintf(&output_params[CXN_PRIORITY][0], "Error+");
          }
        }
      } /* end for each qualifier*/
    } /* end if qualifiers */
    

    /* Next ho addr */
    if (verify & LFAP_FAR_NEXT_HOP_ADDR) {
      IPaddr = *(struct in_addr *)&lfap_far->next_hop_addr.addr_ie.val[0];
      sprintf(&output_params[NEXT_HOP_ADDR][0], "%s", lfap_far->next_hop_addr.addr_ie.family == 1 ? inet_ntoa(IPaddr) : NOT_IP);
    }


    /* Src/Dst VLAN ID */
    if (verify & LFAP_FAR_VLAN_ID) {
      if (lfap_far->vlan_id.i16_ie.val[0] != LFAP_NO_VLAN_ID)
        sprintf(&output_params[SRC_VLAN_ID][0], "%u", lfap_far->vlan_id.i16_ie.val[0]);

      if ( lfap_far->vlan_id.i16_ie.val[1] != LFAP_NO_VLAN_ID )
        sprintf(&output_params[DEST_VLAN_ID][0], "%u",lfap_far->vlan_id.i16_ie.val[1]);
    }
}

/* Gather up the IP specific flow parameters
 */
void
get_ip_output_params(LFAP_FAR *lfap_far, int verify)
{
    unsigned int tos[2];
    unsigned int type[2];
    unsigned short dstPort;
    unsigned char protocol;
    
    /* SRC Port */
    if ( verify & LFAP_FAR_SRC_PORT )
      sprintf(&output_params[SRC_PORT][0], "%lu", lfap_far->src_port.i32_ie.val);
    
    /* Dest port -  part of the protocol IE */
    if ( verify & LFAP_FAR_PROTOCOL_ID ) {
      if (getDestPortLfapFAR(lfap_far, &dstPort) == LFAP_SUCCESS ) {
        if ( dstPort == 0 )
          sprintf(&output_params[DEST_PORT][0], "%s", "none");
        else
          sprintf(&output_params[DEST_PORT][0], "%u", dstPort);
      }
      else
        sprintf(&output_params[DEST_PORT][0], "Error"); /* Why isn't it there?? */
    }

    /* Protocol (TCP/UDP) */
    if ( verify & LFAP_FAR_PROTOCOL_ID ) {
      if ( getIPprotocolLfapFAR(lfap_far, &protocol) == LFAP_SUCCESS ) {
        sprintf(&output_params[PROTOCOL][0],"%u(%s)", protocol, pro2str(protocol));
      }
      else
        sprintf(&output_params[PROTOCOL][0],"Error");
    }

    
    /* Class of Service (TOS) */
    if ( verify & LFAP_FAR_COS ) {
      /* Only print the received and transmitted IPV4 TOS value */
      type[0] = lfap_far->CoS_Arr[LFAP_COS_TYPE_IPV4][LFAP_COS_RECEIVED].CoS.CoS_type;
      tos[0] = lfap_far->CoS_Arr[LFAP_COS_TYPE_IPV4][LFAP_COS_RECEIVED].CoS.val;
      
      type[1] = lfap_far->CoS_Arr[LFAP_COS_TYPE_IPV4][LFAP_COS_TRANSMITTED].CoS.CoS_type;
      tos[1] = lfap_far->CoS_Arr[LFAP_COS_TYPE_IPV4][LFAP_COS_TRANSMITTED].CoS.val;

      if ( type[0] )
        sprintf(&output_params[COS_ARR][0], "%u, ", tos[0]);
      else
        sprintf(&output_params[COS_ARR][0], "none, ");
        

      if ( type[1] )
        sprintf(&output_params[COS_ARR][strlen(&output_params[COS_ARR][0])-1], "%u", tos[1]);
      else
        sprintf(&output_params[COS_ARR][strlen(&output_params[COS_ARR][0])-1], "none");
        
    }      

    /* SRC AS */
    if ( verify & LFAP_FAR_SRC_AS )
      sprintf(&output_params[SRC_AS][0], "%lu", lfap_far->src_AS.i32_ie.val);

      
    /* DST AS */
    if ( verify & LFAP_FAR_DEST_AS  )
      sprintf(&output_params[DEST_AS][0], "%lu", lfap_far->dest_AS.i32_ie.val);

}

/* Gather up the MPLS specific flow parameters
 */
void

get_mpls_output_params(LFAP_FAR *lfap_far, int verify)
{

  int i;
  char tmp[MAX_PARAM_SIZE];
  
  sprintf(&output_params[IS_MPLS][0], "MPLS");

  /* IN LABELS */
  sprintf(&output_params[IN_MPLS_LABEL][0],"{");
  for(i = 0; i < lfap_far->inLabelCount; i++) {
    if ( verify & LFAP_FAR_IN_MPLS_LABEL ) {
      sprintf(tmp, "%u", lfap_far->inMplsLabel[i].mplsLabel.val);
      strcat(&output_params[IN_MPLS_LABEL][0], tmp);
    }
    else
      strcat(&output_params[IN_MPLS_LABEL][0], "0");

    if ((i+1) < lfap_far->inLabelCount)
      strcat(&output_params[IN_MPLS_LABEL][0], ", ");
  }
  strcat(&output_params[IN_MPLS_LABEL][0], "}");
  
  /* OUT LABELS */
  sprintf(&output_params[OUT_MPLS_LABEL][0], "{");
  for(i = 0; i < lfap_far->outLabelCount; i++) {
    if ( verify & LFAP_FAR_OUT_MPLS_LABEL ) {
    sprintf(tmp, "%u", lfap_far->outMplsLabel[i].mplsLabel.val);
    strcat(&output_params[OUT_MPLS_LABEL][0], tmp);
    }
    else
      strcat(&output_params[OUT_MPLS_LABEL][0], "0");

    if ((i+1) < lfap_far->outLabelCount)
      strcat(&output_params[OUT_MPLS_LABEL][0], ", ");
  }
  strcat(&output_params[OUT_MPLS_LABEL][0], "}");
    

  /* FEC */
  sprintf(&output_params[LDP_FEC][0], "{");
  if (verify & LFAP_FAR_LDP_FEC) {
    struct in_addr IPaddr;
    unsigned long tmp_addr;
    
    for(i = 0; i < LFAP_MAX_FECs; i++) {
      switch (lfap_far->ldpFEC[i].ldp.fec.fecType) {
        
      case LFAP_LDP_PREFIX_FEC:
        tmp_addr = htonl(lfap_far->ldpFEC[i].ldp.fec.prefix.addr);
        IPaddr = *(struct in_addr*)&tmp_addr;
        sprintf(tmp, "%s/%d ",  inet_ntoa(IPaddr), lfap_far->ldpFEC[i].ldp.fec.prefix.len);
        strcat(&output_params[LDP_FEC][0], tmp);
        break;
        
      case LFAP_LDP_HOST_FEC:
        tmp_addr = htonl(lfap_far->ldpFEC[i].ldp.fec.host.addr[0]);
        IPaddr = *(struct in_addr*)&tmp_addr;
        sprintf(tmp, "%s/32 ",  inet_ntoa(IPaddr));
        strcat(&output_params[LDP_FEC][0], tmp);
        break;
        
      case LFAP_LDP_VC_FEC:
        if (lfap_far->ldpFEC[i].ldp.fec.vc.type == LFAP_LDP_VC_ETHERNET_VLAN) {
          if (lfap_far->ldpFEC[i].ldp.fec.vc.groupID == 0) {
            /* There is no customer ID  - VCID contains VLAN */
            sprintf(tmp, "0, %lu ", lfap_far->ldpFEC[i].ldp.fec.vc.VCID);
            strcat(&output_params[LDP_FEC][0], tmp);
          }
          else {
            /* group ID contains customer ID and VCID contains the VLAN */
            sprintf(tmp, "%lu, %lu ", lfap_far->ldpFEC[i].ldp.fec.vc.groupID, lfap_far->ldpFEC[i].ldp.fec.vc.VCID);
            strcat(&output_params[LDP_FEC][0], tmp);
          }
        }
        else if (lfap_far->ldpFEC[i].ldp.fec.vc.type == LFAP_LDP_VC_ETHERNET) {
          /* VCID contains customer ID and there is no VLAN */
          sprintf(tmp, "%lu, 0 ", lfap_far->ldpFEC[i].ldp.fec.vc.VCID);
          strcat(&output_params[LDP_FEC][0], tmp);
        }
        else
          /* Unexpected VC type. note - others are defined but just not coded yet */
          strcat(&output_params[LDP_FEC][0], "Unexpected VC, Unexpected VC ");
        break;
        
      default:
        strcat(&output_params[LDP_FEC][0], "Unknown FEC type ");
      }
    }
    if (i+1 < lfap_far->fecCount)
      strcat(&output_params[LDP_FEC][0], ", ");
  }
  strcat(&output_params[LDP_FEC][0], "}");
  
}


/* output to stdout a far record
 */
void
output_far(LFAP_FAR *lfap_far, int verify)
{
    unsigned int i;
    unsigned long flow_type;
    unsigned long first = 0;
    get_common_output_params(lfap_far, verify);
    getFlowTypeLfapFAR(lfap_far, &flow_type);

    if ( flow_type == LFAP_IP ) {
      get_ip_output_params(lfap_far, verify);
      /* See the enum list at the top of the file for output order info 
       * or use   sfas -h   
       *
       */
      for(i = 0; i < MAX_PARAMS; ++i) {
        if ( output_flag[i][IP_FLAG] ) {
          if ( first++ ) printf(", ");
          printf("%s", output_params[i]);
        }
      }
      printf("\n");
    }
    else {
      get_mpls_output_params(lfap_far, verify);
      /* See the enum list at the top of the file for output order info 
       * or use   sfas -h   
       *
       */
      for(i = 0; i < MAX_PARAMS; ++i) {
        if ( output_flag[i][MPLS_FLAG] ) {
          if ( first++ ) printf(", ");
          printf("%s", output_params[i]);
        }
      }
      printf("\n");
    }
      
}



void
process_far(char *buf, int buf_len)
{
     int i;
     unsigned int verify;
     LFAP_FAR lfap_far;
     
     stats.farc++;

     if (the_fas_state != SEND && !loose_handshake) {
       printf("Got FAR message before handhake complete\n");
       the_fas_state = NOT_CONNECTED;
       stats.session_establishment_error++;
       return;
     }


     initRcvdLfapFAR( &lfap_far, buf, buf_len );
     if ( (verify = verifyLfapFAR(&lfap_far)) != LFAP_FAR_INVALID ) {
         if (sfas_debug ) PRINT_LFAP_MESSAGE(&lfap_far, 0, 0);

         for( i = 0; i < MAX_PARAMS; ++i) {
           strcpy(output_params[i], "none");
         }
         output_far(&lfap_far, verify);
     }
     else {
       printf("Got bad FAR msg\n");
       the_fas_state = NOT_CONNECTED;
       return;
     }
}


char *
state2label(unsigned state)
{
    switch (state) {
    case LFAP_INACTIVE:
        return "inactive";        


    case LFAP_ACTIVE:
        return "active";
        
    default:
        return "None";
    }
}


/*
 * print out whatever is in the fun msg...
 */
 
void
output_fun(LFAP_FUN *fun, unsigned int verify)
{
#ifdef _WIN32
    char *byteStr = "(%s) bytes rx %I64u, tx %I64u";
    char *pktStr = "(%s) pkts rx %I64u, tx %I64u";
#else
    char *byteStr = "(%s) bytes rx %llu,  tx %llu";
    char *pktStr = "(%s) pkts rx %llu,  tx %llu";
#endif
    int i;
    unsigned long first = 0;
    longlong_t rcvd;
    longlong_t sent;
    char *d_t;
    char *ti;
    time_t t;
    
    /* FLow ID */
    sprintf(&output_params[FLOW_ID][0], "U:%lu.%lu.%lu",
           fun->flow_id.fid_ie.fas_id[0], 
           fun->flow_id.fid_ie.fas_id[1], 
            fun->flow_id.fid_ie.cce_id);

    /* Update time */
    t = ara_time + ((fun->time.i32_ie.val - switch_time) / SWITCH_TICS_PER_SEC);
    ti = ctime(&t);
    ti[strlen(ti) - 1] = 0;
    sprintf(&output_params[TIME][0], "%s", ti);
    
    
    /* Flow STate */
    if (verify & LFAP_FUN_FLOW_STATE)
        sprintf(&output_params[FUN_STATE][0],"%u(%s)", fun->flow_state.rsvd_ui8.val, state2label(fun->flow_state.rsvd_ui8.val));
    else
        sprintf(&output_params[FUN_STATE][0],"%u(%s)",  LFAP_ACTIVE, state2label(LFAP_ACTIVE));
    
    
    /* Byte counmt */
    if (!(verify & LFAP_FUN_BYTE_COUNT))
        sprintf(&output_params[BYTES][0],"none, none");
    else {
        rcvd = ((longlong_t) fun->byte_count.dataCount_ie.dc_rcvd.hi_val << 32 ) |
	       fun->byte_count.dataCount_ie.dc_rcvd.lo_val;
	sent = ((longlong_t)fun->byte_count.dataCount_ie.dc_sent.hi_val << 32 ) |
	       fun->byte_count.dataCount_ie.dc_sent.lo_val;
        d_t = fun->byte_count.dataCount_ie.type == LFAP_BYTES_DELTA ? "delta" : "absolute",
            sprintf(&output_params[BYTES][0],byteStr, d_t, rcvd, sent);
    }
    

    /* Packet count */
    if (!(verify & LFAP_FUN_PACKET_COUNT))
        sprintf(&output_params[PKTS][0],"none, none");
    else {
        
        rcvd = ((longlong_t) fun->packet_count.dataCount_ie.dc_rcvd.hi_val << 32 ) |
	       fun->packet_count.dataCount_ie.dc_rcvd.lo_val;
	sent = ((longlong_t)fun->packet_count.dataCount_ie.dc_sent.hi_val << 32 ) |
	       fun->packet_count.dataCount_ie.dc_sent.lo_val;
        d_t = fun->packet_count.dataCount_ie.type == LFAP_BYTES_DELTA ? "delta" : "absolute",
            sprintf(&output_params[PKTS][0], pktStr,  d_t, rcvd, sent);
    }
    
    /* See the enum list at the top of the file for output order info 
     * or use   sfas -h   
     *
     */
    for(i = 0; i < MAX_PARAMS; ++i) {
      if ( output_flag[i][FUN_FLAG] ) {
        if ( first++ ) printf(", ");
        printf("%s", output_params[i]);
      }
    }
    printf("\n");
}

void
process_fun(char *buf, int buf_len)
{
    int i;
    LFAP_FUN funStruct;
    unsigned int verify;
    
    stats.fun++; /* we count the number of messages not flows */
                 /* so leave it outside the while loop.       */


    if (the_fas_state != SEND && !loose_handshake) {
      printf("Got FUN message before handhake complete\n");
      the_fas_state = NOT_CONNECTED;
      stats.session_establishment_error++;
      return;
    }


    initRcvdLfapFUN( &funStruct, buf, buf_len );
    verify = verifyLfapFUN(&funStruct);
    do {
      if (verify != LFAP_FUN_INVALID) {
        if (sfas_debug) PRINT_LFAP_MESSAGE(&funStruct, 0, 0);

        for( i = 0; i < MAX_PARAMS; ++i) {
          strcpy(output_params[i], "none");
        }
        output_fun(&funStruct, verify);
      }
      else {

          printf("Got bad FUN msg\n");
          the_fas_state = NOT_CONNECTED;
        return;
      }
    } while ( decodeNextRecordLfapFUN( &funStruct ) != LFAP_EMPTY_BUFFER );
    
}

/* pre: recieved a message from a CCE
 * post: process data and optionally send response to CCE  * on error close connection to CCE, output error msg
 */
void
process_lfap_msg(char *msg, int msgLen, SOCKET client)
{
  unsigned char opc;


    opc = lfap_msg_get_opcode(msg);
    SHOW_DEBUG_MSG1("Processing message %s", op_strings[opc]);
    
    if ( lfap_msg_get_version(msg) != LFAP_VERSION  ) {
      /* According to the spec the only time it is OK to receive a 
       * message with a differnet version number is on the initial 
       * VR message. The check is best expressed as a positive.
       */
        if (the_fas_state == CONNECT && opc == LFAP_VR_OPCODE) {
        /* OK, allowed */
        }
        else {
          printf("Got message with bad version %d\n", lfap_msg_get_version(msg));
          dump_buffer(msg, msgLen);
          stats.bad_version_rx++;            
          the_fas_state = NOT_CONNECTED;
          return;
        }
    }
    
    switch (opc) {
        
    case LFAP_VR_OPCODE:
        process_vr(client, msg, msgLen);
        break;
        
    case LFAP_VRA_OPCODE:
        SHOW_DEBUG_MSG("got an unexpected VRA msg from CCE");
        stats.unexpected_vra++;
        the_fas_state = NOT_CONNECTED;
        break;
        
    case LFAP_CR_OPCODE:
        process_cr(client, msg, msgLen);
        send_ar_get_time(client);
        break;

    case LFAP_FAR_OPCODE:
        process_far(&msg[0], msgLen);
        break;
        
    case LFAP_FUN_OPCODE:
        process_fun(&msg[0], msgLen);
        break;
        
    case LFAP_AR_OPCODE:
        process_ar(client, msg, msgLen);
        break;

    case LFAP_ARA_OPCODE:
        process_ara( client, msg, msgLen );
        break;

    case LFAP_CAN_OPCODE:
    case LFAP_CRN_OPCODE:
    case LFAP_FER_OPCODE:
    case LFAP_KA_OPCODE:
    case LFAP_DR_OPCODE:
        /* should not receive */
        printf("Unexpected message received %s\n", op_strings[opc]);
        break;

    default:
        printf("process_lfap_msg: error: unknown opcode: %d len:%d\n", opc, msgLen);
        dump_buffer(msg, msgLen);
        the_fas_state = NOT_CONNECTED;        
        SHOW_DEBUG_MSG1("process_lfap_msg: invalid opcode %d, igonoring", opc);
        stats.bad_opcode++;
    }

}


void
run_lfap_protocol(SOCKET client)
{
    char lfapMsg[LFAP_BUFFER_SIZE];
    int rc;
    unsigned short len;
    fd_set readfds;
    struct timeval tv = {keep_alive_interval, 0};
    time_t last_time, curr_time;


    SHOW_DEBUG_MSG1("run_lfap_protocol: new client on sd %d", client);
    
    last_time = time(NULL);
    
    while (the_fas_state != NOT_CONNECTED ) {

        /* 0. wait for timeout or msg from cce (protocol does not require this though
           but ssr implementation will timeout if it does not get some msgs) */


        FD_ZERO(&readfds);
        FD_SET(client, &readfds);        
        rc = select(client +1, &readfds, NULL, NULL, &tv);
        if (rc == -1) {
            perror("run_lfap_protocol: select");
            return;
        }
        
        /* timeout - need to send msg to cce to prevent timeout */
        if (the_fas_state == SEND) {
          curr_time = time(NULL);
          if (rc == 0 || difftime(curr_time, last_time) > keep_alive_interval) {
            send_keepalive(client);
            tv.tv_sec = keep_alive_interval;
            last_time = curr_time;
          }
        }
        
        /* message from client */
        if (rc > 0 && FD_ISSET(client, &readfds)) {
            memset(&lfapMsg[0], 0, sizeof(lfapMsg)); /* good for debugging, but not necessary */
            /* tcp_readn returns exactly the size requested */
            rc = tcp_readn(client, (char *)&lfapMsg[0], LFAP_HEADER_SIZE);
            if (rc == -1) {
                SHOW_DEBUG_MSG("run_lfap_protocol: read failed, session lost");
                stats.bad_read++;
                return;
             }
            
        /* 2. verify we have a message */
            if (rc == 0) {
                printf("run_lfap_protocol: remote client sd=%d dropped connection\n", client);
                return;
            }
            
            stats.msgs++;
            stats.bytes += rc;
            
            /* 3. parse var len */
            len = lfap_msg_get_msglen(lfapMsg);


            /* 4. load the remainder of the message into the buffer */
            if (len > 0) {
                rc = tcp_readn(client, &lfapMsg[LFAP_HEADER_SIZE], len);
                if (rc == -1) {
                    SHOW_DEBUG_MSG("read failed, session lost");
                stats.bad_read++;
                return;
                }
                stats.bytes += rc;            
            }
            
            /* 5. process msg */
            if (rc > 0)
              /* Note the header size is not included in the message 
               * length so add it in to get the total size 
               */
                process_lfap_msg(&lfapMsg[0], len + LFAP_HEADER_SIZE, client);
            else 
                return;
        }
    }
}

void
process_clients(SOCKET sd)
{
    SOCKET client;
    struct sockaddr_in from;
    int addrlen = sizeof(from);
    time_t clock;
    
    time(&clock);
    
    while (1) {
        if ( ( client = accept ( sd, (struct sockaddr *)&from, &addrlen) ) < 0 ) {
            SHOW_DEBUG_MSG1("network server accept failure %d", errno);
            perror("network server");
            return;
        }
        
        stats.connection_count++;
        printf("CCE: %s connected on %s", inet_ntoa(from.sin_addr), ctime(&clock));
        SHOW_DEBUG_MSG1("connection count at %d", connection_count);
        the_fas_state = CONNECT;
        ar_time_msg_id = 0;
        run_lfap_protocol(client);
        closesocket(client);
        stats.lost_session++;
        the_fas_state = NOT_CONNECTED;
        printf("CCE: lost session with %s on %s", inet_ntoa(from.sin_addr), ctime(&clock));
    }
}


/* pre: os sent signal to app
 * post: output statistics, terminate app
 */
void sig_handler(int id)
{
    printf("\nCaught signal %d Exiting\n", id);

    printf("Stats:\
bad opcode %d bad handhshake %d unexpected ara %d unexpected vra %d\n\
bad read %d bad VR version received %d \n\
lost session %d number connections %d\n\
messages %d bytes %d\n",
    stats.bad_opcode,
    stats.session_establishment_error,
    stats.unexpected_ara,
    stats.unexpected_vra,    
    stats.bad_read,
    stats.bad_version_rx,
    stats.lost_session,
    stats.connection_count,
    stats.msgs,
    stats.bytes);


#ifdef _WIN32
    WSACleanup();
#endif
        
    exit(id);
}

static void
usage()
{
  int i;

  printf("\n\nusage: sfas [-s enable strict handshake -d enable diagnostics]\n\n");
  printf("Two types of records are printed FAR(F:) and FUN (U:). \n\n\
A FAR for IP flows consists of the following fields, optional fields will have a \"none\" value\n\n\
F                     : far msg type\n");

  for(i = 0; i < MAX_PARAMS; ++i) {
    if ( output_flag[i][IP_FLAG] )
      printf("%s\n", output_strings[i]);
  }

  printf("\n\nA FAR for MPLS flows consists of the following fields, optional fields will have a \"none\" value\n\n\
F                     : far msg type\n");
  for(i = 0; i < MAX_PARAMS; ++i) {
    if ( output_flag[i][MPLS_FLAG] )
      printf("%s\n", output_strings[i]);
  }

  
  printf("\nA FUN consists of the following fields:\n\
U                     : fun msg type\n");

  for(i = 0; i < MAX_PARAMS; ++i) {
    if ( output_flag[i][FUN_FLAG] )
      printf("%s\n", output_strings[i]);
  }

}
extern void  lfap_init_encode();
extern void   lfap_init_decode();


int
main(int argc, char *argv[])
{

  int i;
  int rc = 0;
  SOCKET sd = 0;
#ifdef _WIN32
  WORD wsv = MAKEWORD(1,1);
  WSADATA wsa;
  if (WSAStartup(wsv, &wsa)) {
    printf("Windows can't load winsock dll. Install TCP/IP Exiting\n");
  }
#endif

  /* do some LFAP setup */
  lfap_init_encode();
  lfap_init_decode();

  signal(SIGINT, sig_handler);

  for(i = 0; i <= LFAP_MAX_OPCODE; i++) 
    op_strings[i] = "invalid opcode";

  op_strings[LFAP_VR_OPCODE]    = "LFAP_VR_OPCODE";
  op_strings[LFAP_VRA_OPCODE]   = "LFAP_VRA_OPCODE";
  op_strings[LFAP_CR_OPCODE]    = "LFAP_CR_OPCODE";
  op_strings[LFAP_CAN_OPCODE]   = "LFAP_CAN_OPCODE";
  op_strings[LFAP_CRN_OPCODE]   = "LFAP_CRN_OPCODE";
  op_strings[LFAP_FER_OPCODE]   = "LFAP_FER_OPCODE";
  op_strings[LFAP_FAR_OPCODE]   = "LFAP_FAR_OPCODE";
  op_strings[LFAP_FUN_OPCODE]   = "LFAP_FUN_OPCODE";
  op_strings[LFAP_AR_OPCODE]    = "LFAP_AR_OPCODE";
  op_strings[LFAP_ARA_OPCODE]   = "LFAP_ARA_OPCODE";
  op_strings[LFAP_KA_OPCODE]    = "LFAP_KA_OPCODE";
  op_strings[LFAP_DR_OPCODE]    = "LFAP_DR_OPCODE";


  for(i = 0; i <= LFAP_MAX_CMD_CODE; i++)
    op_cmd_codes[i] = "invalid cmd code";

  op_cmd_codes[LFAP_RETURN_INDICATED_FLOWS] = "LFAP_RETURN_INDICATED_FLOWS";
  op_cmd_codes[LFAP_RETURN_FLOW_PREFIX]     = "LFAP_RETURN_FLOW_PREFIX";
  op_cmd_codes[LFAP_RETURN_TIME]            = "LFAP_RETURN_TIME";
  op_cmd_codes[LFAP_REQUESTED_IEs]          = "LFAP_REQUESTED_IES";


  for(i = 0; i <= LFAP_MAX_IE_CODE; i++)
    ie_codes[i] = "invalid IE type";

  ie_codes[LFAP_TRANSPORT_MULTIPLE_RECORD] = "LFAP_TRANSPORT_MULTIPLE_RECORD";
  ie_codes[LFAP_TRANSPORT_FAS_IP]     = "LFAP_TRANSPORT_FAS_IP";

  ie_codes[LFAP_FLOW_ID]         = "LFAP_FLOW_ID";
  ie_codes[LFAP_SRC_ADDR]        = "LFAP_SRC_ADDR";
  ie_codes[LFAP_DST_ADDR]        = "LFAP_DST_ADDR";
  ie_codes[LFAP_FLOW_QUALIFIER]  = "LFAP_FLOW_QUALIFIER";
  ie_codes[LFAP_TIME]            = "LFAP_TIME";
  ie_codes[LFAP_CLASS_OF_SERVICE] = "LFAP_CLASS_OF_SERVICE";
  ie_codes[LFAP_SRC_CCE_ADDR]    = "LFAP_SRC_CCE_ADDR";
  ie_codes[LFAP_CLIENT_DATA]     = "LFAP_CLIENT_DATA";
  ie_codes[LFAP_COMMAND_CODE]    = "LFAP_COMMAND_CODE";
  ie_codes[LFAP_FAS_IP_ADDR]     = "LFAP_FAS_IP_ADDR";
  ie_codes[LFAP_FAIL_CODE]       = "LFAP_FAIL_CODE";
  ie_codes[LFAP_FLOW_ID_PREFIX]  = "LFAP_FLOW_ID_PREFIX";
  ie_codes[LFAP_FLOW_STATE]      = "LFAP_FLOW_STATE";
  ie_codes[LFAP_BYTES_TOTAL]     = "LFAP_BYTES_TOTAL";
  ie_codes[LFAP_BYTES_DELTA]     = "LFAP_BYTES_DELTA";
  ie_codes[LFAP_PACKETS_TOTAL]   = "LFAP_PACKETS_TOTAL";
  ie_codes[LFAP_PACKETS_DELTA]   = "LFAP_PACKETS_DELTA";
  ie_codes[LFAP_PROTOCOL_ID]     = "LFAP_PROTOCOL_ID";
  ie_codes[LFAP_SRC_PORT_UDP]    = "LFAP_SRC_PORT_UDP";
  ie_codes[LFAP_SRC_PORT_TCP]    = "LFAP_SRC_PORT_TCP";
  ie_codes[LFAP_SRC_SOCKET_IPX]  = "LFAP_SRC_SOCKET_IPX";
  ie_codes[LFAP_SRC_AS]          = "LFAP_SRC_AS";
  ie_codes[LFAP_INGRESS_PORT]    = "LFAP_INGRESS_PORT";
  ie_codes[LFAP_EGRESS_PORT]     = "LFAP_EGRESS_PORT";


  for(i = 0; i < MAX_PARAMS; ++i) {

    switch (i) {
    case FLOW_ID:
      output_flag[FLOW_ID][IP_FLAG]          = 1;
      output_flag[FLOW_ID][MPLS_FLAG]        = 1;
      output_flag[FLOW_ID][FUN_FLAG]         = 1;
      output_strings[FLOW_ID]                = "Flow identifier       : 3 integers separated by a period aka dot to uniquely identify flow";
      break;
    case TIME:
      output_flag[TIME][IP_FLAG]             = 1;
      output_flag[TIME][MPLS_FLAG]           = 1;
      output_flag[TIME][FUN_FLAG]            = 1;
      output_strings[TIME]                   = "sysUpTime             : switch timestamp when flow was learned or updated by the switch";
      break;
    case SRC_ADDR:
      output_flag[SRC_ADDR][IP_FLAG]         = 1;
      output_flag[SRC_ADDR][MPLS_FLAG]       = 1;
      output_flag[SRC_ADDR][FUN_FLAG]        = 0;
      output_strings[SRC_ADDR]               = "Source IP             : IPv4 source address";
      break;
    case DEST_ADDR:
      output_flag[DEST_ADDR][IP_FLAG]        = 1;
      output_flag[DEST_ADDR][MPLS_FLAG]      = 1;
      output_flag[DEST_ADDR][FUN_FLAG]       = 0;
      output_strings[DEST_ADDR]              = "Dest IP               : IPv4 destination address";
      break;
    case SRC_PORT:
      output_flag[SRC_PORT][IP_FLAG]         = 1;
      output_flag[SRC_PORT][MPLS_FLAG]       = 0;
      output_flag[SRC_PORT][FUN_FLAG]        = 0;
      output_strings[SRC_PORT]               = "Source port           : if packet is TCP or UDP, the source port number";
      break;
    case DEST_PORT:
      output_flag[DEST_PORT][IP_FLAG]        = 1;
      output_flag[DEST_PORT][MPLS_FLAG]      = 0;
      output_flag[DEST_PORT][FUN_FLAG]       = 0;
      output_strings[DEST_PORT]              = "Dest port             : if packet is TCP or UDP, the destination port number";
      break;
    case PROTOCOL:
      output_flag[PROTOCOL][IP_FLAG]         = 1;
      output_flag[PROTOCOL][MPLS_FLAG]       = 0;
      output_flag[PROTOCOL][FUN_FLAG]        = 0;
      output_strings[PROTOCOL]               = "IP protocol           : The ipv4 packet header protocol value, example: 17(udp)";
      break;
    case SRC_CCE_ADDR:
      output_flag[SRC_CCE_ADDR][IP_FLAG]     = 1;
      output_flag[SRC_CCE_ADDR][MPLS_FLAG]   = 1;
      output_flag[SRC_CCE_ADDR][FUN_FLAG]    = 0;
      output_strings[SRC_CCE_ADDR]           = "Originating Switch    : Switch accounting for this flow (useful in proxy setups)";
      break;
    case SRC_AS:
      output_flag[SRC_AS][IP_FLAG]           = 1;
      output_flag[SRC_AS][MPLS_FLAG]         = 0;
      output_flag[SRC_AS][FUN_FLAG]          = 0;
      output_strings[SRC_AS]                 = "Src Autonomus System  : AS number";
      break;
    case INGRESS_PORT:
      output_flag[INGRESS_PORT][IP_FLAG]     = 1;
      output_flag[INGRESS_PORT][MPLS_FLAG]   = 1;
      output_flag[INGRESS_PORT][FUN_FLAG]    = 0;
      output_strings[INGRESS_PORT]           = "Ingress ifIndex       : The switch port identified by fIndex (see SNMP ifTable)";
      break;
    case EGRESS_PORT:
      output_flag[EGRESS_PORT][IP_FLAG]      = 1;
      output_flag[EGRESS_PORT][MPLS_FLAG]    = 1;
      output_flag[EGRESS_PORT][FUN_FLAG]     = 0;
      output_strings[EGRESS_PORT]            = "Egress ifIndex        : The switch port identified by fIndex (see SNMP ifTable)";
      break;
    case SRC_VLAN_ID:
      output_flag[SRC_VLAN_ID][IP_FLAG]      = 1;
      output_flag[SRC_VLAN_ID][MPLS_FLAG]    = 1;
      output_flag[SRC_VLAN_ID][FUN_FLAG]     = 0;
      output_strings[SRC_VLAN_ID]            = "Src VLAN ID           : VLAN ID of the ingress port";
      break;
    case DEST_VLAN_ID:
      output_flag[DEST_VLAN_ID][IP_FLAG]     = 1;
      output_flag[DEST_VLAN_ID][MPLS_FLAG]   = 1;
      output_flag[DEST_VLAN_ID][FUN_FLAG]    = 0;
      output_strings[DEST_VLAN_ID]           = "Dest VLAN ID          : VLAN ID of the egress port";
      break;
    case IN_MPLS_LABEL:
      output_flag[IN_MPLS_LABEL][IP_FLAG]    = 0;
      output_flag[IN_MPLS_LABEL][MPLS_FLAG]  = 1;
      output_flag[IN_MPLS_LABEL][FUN_FLAG]   = 0;
      output_strings[IN_MPLS_LABEL]          = "Incoming Label Stack  : Up to 4 MPLS labels; listed inside {}";
      break;
    case OUT_MPLS_LABEL:
      output_flag[OUT_MPLS_LABEL][IP_FLAG]   = 0;
      output_flag[OUT_MPLS_LABEL][MPLS_FLAG] = 1;
      output_flag[OUT_MPLS_LABEL][FUN_FLAG]  = 0;
      output_strings[OUT_MPLS_LABEL]         = "Outgoing Label Stack  : Up to 4 MPLS labels; listed inside {}";
      break;
    case LDP_FEC:
      output_flag[LDP_FEC][IP_FLAG]          = 0;
      output_flag[LDP_FEC][MPLS_FLAG]        = 1;
      output_flag[LDP_FEC][FUN_FLAG]         = 0;
      output_strings[LDP_FEC]                = "LDP FEC               : The FEC for MPLS flows set up with LDP; listed inside {}";
      break;
    case COS_ARR:
      output_flag[COS_ARR][IP_FLAG]         = 1;
      output_flag[COS_ARR][MPLS_FLAG]       = 0;
      output_flag[COS_ARR][FUN_FLAG]        = 0;
      output_strings[COS_ARR]               =  "DSCP(TOS)             : The incoming ipv4 packet headers DSCP byte\nDSCP(TOS)             : The outgoing ipv4 packet headers DSCP bye";
      break;
    case NEXT_HOP_ADDR:
      output_flag[NEXT_HOP_ADDR][IP_FLAG]    = 1;
      output_flag[NEXT_HOP_ADDR][MPLS_FLAG]  = 1;
      output_flag[NEXT_HOP_ADDR][FUN_FLAG]   = 0;
      output_strings[NEXT_HOP_ADDR]          = "Next Hop Addr         : The IP address of the next hop";
      break;
    case DEST_AS:
      output_flag[DEST_AS][IP_FLAG]          = 1;
      output_flag[DEST_AS][MPLS_FLAG]        = 0;
      output_flag[DEST_AS][FUN_FLAG]         = 0;
      output_strings[DEST_AS]                = "Dst Autonomus System  : AS number";
      break;
    case CXN_PRIORITY:
      output_flag[CXN_PRIORITY][IP_FLAG]     = 1;
      output_flag[CXN_PRIORITY][MPLS_FLAG]   = 1;
      output_flag[CXN_PRIORITY][FUN_FLAG]    = 0;
      output_strings[CXN_PRIORITY]           = "Service Priority queue: Which internal switch priority queue the flow exists in";
      break;
    case CHECKPOINT:
      output_flag[CHECKPOINT][IP_FLAG]       = 1;
      output_flag[CHECKPOINT][MPLS_FLAG]     = 1;
      output_flag[CHECKPOINT][FUN_FLAG]      = 0;
      output_strings[CHECKPOINT]             = "Checkpoint interval   : Not used by slate";
      break;
    case IS_MPLS:
      output_flag[IS_MPLS][IP_FLAG]          = 0;
      output_flag[IS_MPLS][MPLS_FLAG]        = 1;
      output_flag[IS_MPLS][FUN_FLAG]         = 0;
      output_strings[IS_MPLS]                = "MPLS                  : Indicates this is an MPLS flow";
      break;
    case FUN_STATE:
      output_flag[FUN_STATE][IP_FLAG]        = 0;
      output_flag[FUN_STATE][MPLS_FLAG]      = 0;
      output_flag[FUN_STATE][FUN_FLAG]       = 1;
      output_strings[FUN_STATE]              = "Flow state            : current flow status (active or inactive)";
      break;
    case BYTES:
      output_flag[BYTES][IP_FLAG]            = 0;
      output_flag[BYTES][MPLS_FLAG]          = 0;
      output_flag[BYTES][FUN_FLAG]           = 1;
      output_strings[BYTES]                  = "Byte  statistics      : Bytes sent and received (64 bit integers) absolute or delta";
      break;
    case PKTS:
      output_flag[PKTS][IP_FLAG]             = 0;
      output_flag[PKTS][MPLS_FLAG]           = 0;
      output_flag[PKTS][FUN_FLAG]            = 1;
      output_strings[PKTS]                   = "Packets statistics    : Packets sent and received (64 bit integers) absolute or deltaPKTS";
      break;
   default:
      output_flag[i][IP_FLAG]                = 1;
      output_flag[i][MPLS_FLAG]              = 1;
      output_flag[i][FUN_FLAG]               = 1;
      output_strings[i]                      = "Error                 : Unexpected field";

    }
  }
    
  printf("\nRiverstone Networks: trivial flow accounting server version %d.%d.%d\n", 
         maj_version, min_version, update_version);
  printf("This server supports only LFAP Version 5\n");
  printf("Use   sfas -h   for help on record format\n\n");
  
  for (rc = 1; rc < argc; rc++) {
      
      if (argv[rc][1] == 's') {
          printf("%s Strict handshake enabled\n", argv[0]);
          loose_handshake = 0;
      }
      
      if (argv[rc][1] == 'd') {
          printf("%s: diagnostic trace enabled\n", argv[0]);
          sfas_debug = 1;
      }
      
      if (argv[rc][1] == 'h') {
          usage();
          goto done;
      }
  }
  
  
  /* 1. open tcp socket for listening */
  sd = create_server();
  
  if (sd == -1) {
      printf("open socket failed: %d %s\n", errno, strerror(errno));
      rc = 1;
      goto done;
  }
  
  
  /* run main loop */
  process_clients(sd);
  
done:
  /* clean up */
  closesocket(sd);
#ifdef _WIN32
  WSACleanup();
#endif
  return rc;
}

